#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import aiohttp
import asyncio
import binascii
import codecs
import logging
from gi.repository import Gtk, Gdk, GdkPixbuf
from typing import Any

from stlib import login, universe
from . import utils
from .. import i18n, config

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class LoginDialog(Gtk.Dialog):
    def __init__(
            self,
            parent_window: Gtk.Window,
            application: Gtk.Application,
            mobile_login: bool = True,
    ) -> None:
        super().__init__(use_header_bar=True)
        self.application = application
        self.mobile_login = mobile_login
        self.has_user_data = False

        self.header_bar = self.get_header_bar()

        self.login_button = utils.AsyncButton()
        self.login_button.set_label(_("Log-in"))
        self.login_button.connect("clicked", self.on_login_button_clicked)
        self.header_bar.pack_end(self.login_button)

        self.parent_window = parent_window
        self.set_default_size(400, 100)
        self.set_title(_('Login'))
        self.set_transient_for(parent_window)
        self.set_modal(True)
        self.set_destroy_with_parent(True)
        self.set_resizable(False)

        self.content_area = self.get_content_area()
        self.content_area.set_orientation(Gtk.Orientation.VERTICAL)
        self.content_area.set_spacing(10)
        self.content_area.set_margin_start(10)
        self.content_area.set_margin_end(10)
        self.content_area.set_margin_top(10)
        self.content_area.set_margin_bottom(10)

        self.status = utils.SimpleStatus()
        self.content_area.append(self.status)

        self.user_details_section = utils.Section("login")
        self.content_area.append(self.user_details_section)

        self.username_item = self.user_details_section.new_item("account_name", _("Username:"), Gtk.Entry, 0, 0)

        self.__password_item = self.user_details_section.new_item("_password", _("Password:"), Gtk.Entry, 0, 1)
        self.__password_item.set_visibility(False)
        self.__password_item.set_invisible_char('*')

        self.steam_code_item = self.user_details_section.new_item("_steam_code", _("Steam Code:"), Gtk.Entry, 0, 2)
        self.mail_code_item = self.user_details_section.new_item("_mail_code", _("Mail Code:"), Gtk.Entry, 0, 2)

        self.captcha_gid = -1
        self.captcha_item = self.user_details_section.new_item("_captcha", _("Code:"), Gtk.Image, 0, 3)
        self.captcha_item.set_size_request(140, 140)
        self.captcha_text_item = self.user_details_section.new_item(
            "_captcha_text", _("Captcha Text:"), Gtk.Entry, 0, 4,
        )

        self.save_password_item = self.user_details_section.new_item("_savepwd", _("Save Password:"), Gtk.CheckButton,
                                                                     0, 5)
        self.save_password_item.set_active(True)

        self.advanced_login = utils.ClickableLabel()
        self.advanced_login.set_markup(utils.markup(_("Advanced Login"), font_size='x-small', color='blue'))
        self.advanced_login.set_halign(Gtk.Align.END)
        self.advanced_login.connect("clicked", self.on_advanced_login_clicked)
        self.content_area.append(self.advanced_login)

        self.advanced_login_section = utils.Section("login")
        self.content_area.append(self.advanced_login_section)

        self.identity_secret_item = self.advanced_login_section.new_item(
            'identity_secret',
            _("Identity Secret:"),
            Gtk.Entry,
            0, 0,
        )

        self.shared_secret_item = self.advanced_login_section.new_item(
            'shared_secret',
            _("Shared Secret:"),
            Gtk.Entry,
            0, 1,
        )

        self.connect('response', self.on_quit)

        key_event = Gtk.EventControllerKey()
        key_event.connect('key-released', self.on_key_release_event)
        self.add_controller(key_event)

        self.steam_code_item.hide()
        self.mail_code_item.hide()
        self.captcha_item.hide()
        self.captcha_text_item.hide()
        self.advanced_login_section.hide()

        self.check_login_availability()

    @property
    def username(self) -> str:
        return self.username_item.get_text()

    @property
    def __password(self) -> str:
        return self.__password_item.get_text()

    def set_password(self, encrypted_password: str) -> None:
        try:
            key = codecs.decode(encrypted_password, 'rot13')
            raw = codecs.decode(key.encode(), 'base64')
            self.__password_item.set_text(raw.decode())
        except (binascii.Error, UnicodeError, TypeError):
            log.warning(_("Password decode failed. Trying RAW."))
            self.__password_item.set_text(encrypted_password)

    @property
    def mail_code(self) -> str:
        return self.mail_code_item.get_text()

    @property
    def steam_code(self) -> str:
        return self.steam_code_item.get_text()

    @property
    def captcha_text(self) -> str:
        return self.captcha_text_item.get_text()

    @property
    def shared_secret(self) -> str:
        return self.shared_secret_item.get_text()

    @property
    def identity_secret(self) -> str:
        return self.identity_secret_item.get_text()

    def check_login_availability(self) -> None:
        if not self.username or not self.__password:
            self.login_button.set_sensitive(False)
        else:
            self.login_button.set_sensitive(True)

    def on_quit(self, *args: Any, **kwargs: Any) -> None:
        self.application.main_window.statusbar.set_warning(
            'steamguard',
            _("Login cancelled! Modules will not work correctly!"),
        )
        self.destroy()

    def on_key_release_event(
            self,
            controller: Gtk.EventControllerKey,
            keyval: int,
            keycode: int,
            state: Gdk.ModifierType
    ) -> None:
        self.check_login_availability()

        if keyval == Gdk.KEY_Return:
            if not self.username or not self.__password:
                self.status.error(_("Username or Password is blank!"))
            else:
                self.login_button.emit('clicked')

    async def on_login_button_clicked(self, *args: Any) -> None:
        self.status.info(_("Retrieving user data"))
        self.application.main_window.statusbar.set_warning("steamguard", _("Not logged in"))
        self.username_item.set_sensitive(False)
        self.__password_item.set_sensitive(False)
        self.save_password_item.set_sensitive(False)
        self.login_button.set_sensitive(False)

        _login_session = login.Login.get_session(0)
        _login_session.username = self.username
        _login_session.password = self.__password

        kwargs = {'emailauth': self.mail_code, 'mobile_login': self.mobile_login}

        # no reason to send captcha_text if no gid is found
        if self.captcha_gid != -1:
            kwargs['captcha_text'] = self.captcha_text
            kwargs['captcha_gid'] = self.captcha_gid
            # if login fails for any reason, gid must be unset
            # CaptchaError exception will reset it if needed
            self.captcha_gid = -1

        if not self.shared_secret or not self.identity_secret:
            log.warning(_("No shared secret found. Trying to log-in without two-factor authentication."))
            # self.code_item.show()

        kwargs['shared_secret'] = self.shared_secret
        kwargs['authenticator_code'] = self.steam_code

        self.status.info(_("Logging in"))
        self.captcha_item.hide()
        self.captcha_text_item.hide()
        self.steam_code_item.hide()
        self.mail_code_item.hide()

        try_count = 3

        while True:
            try:
                login_data = await _login_session.do_login(**kwargs)
            except login.MailCodeError:
                self.status.info(_("Write code received by email\nand click on 'Log-in' button"))
                self.mail_code_item.set_text("")
                self.mail_code_item.show()
                self.mail_code_item.grab_focus()
            except login.TwoFactorCodeError:
                if self.shared_secret:
                    if try_count > 0:
                        log.warning(_("Retrying login in 10 seconds"))
                        await asyncio.sleep(10)
                        try_count -= 1
                        continue
                    else:
                        self.status.error(_("shared secret is invalid!"))
                        self.username_item.set_sensitive(True)
                        self.__password_item.set_sensitive(True)
                        self.shared_secret_item.grab_focus()
                        break

                self.status.error(_("Write Steam Code bellow and click on 'Log-in'"))
                self.steam_code_item.set_text("")
                self.steam_code_item.show()
                self.steam_code_item.grab_focus()
            except login.LoginBlockedError:
                self.status.error(_(
                    "Your network is blocked!\n"
                    "It'll take some time until unblocked. Please, try again later\n"
                ))
                self.user_details_section.hide()
                self.advanced_login.hide()
                self.advanced_login_section.hide()
                self.login_button.hide()
                self.set_deletable(True)
            except login.CaptchaError as exception:
                self.status.info(_("Write captcha code as shown bellow\nand click on 'Log-in' button"))
                self.captcha_gid = exception.captcha_gid

                pixbuf_loader = GdkPixbuf.PixbufLoader()
                pixbuf_loader.set_size(140, 50)
                pixbuf_loader.write(exception.captcha)
                pixbuf_loader.close()
                self.captcha_item.set_from_pixbuf(pixbuf_loader.get_pixbuf())

                self.captcha_item.show()
                self.captcha_text_item.set_text("")
                self.captcha_text_item.show()
                self.captcha_text_item.grab_focus()
            except (login.LoginError, AttributeError) as exception:
                log.error(str(exception))
                # self.__password_item.set_text('')
                self.__password_item.grab_focus()
                config.remove('login', 'token')
                config.remove('login', 'token_secure')
                config.remove('login', 'oauth_token')

                self.status.error(':\n'.join(str(exception).split(': ')))
                self.username_item.set_sensitive(True)
                self.__password_item.set_sensitive(True)
                self.__password_item.grab_focus()
            except binascii.Error:
                self.status.error(_("shared secret is invalid!"))
                self.username_item.set_sensitive(True)
                self.__password_item.set_sensitive(True)
                self.shared_secret_item.grab_focus()
            except (aiohttp.ClientError, ValueError):
                self.status.error(_("Check your connection. (server down?)"))
                await asyncio.sleep(15)
                continue
            else:
                new_configs = {"account_name": self.username}

                if "shared_secret" in login_data.auth:
                    new_configs["shared_secret"] = login_data.auth["shared_secret"]
                elif self.shared_secret:
                    new_configs["shared_secret"] = self.shared_secret

                if "identity_secret" in login_data.auth:
                    new_configs['identity_secret'] = login_data.auth['identity_secret']
                elif self.identity_secret:
                    new_configs["identity_secret"] = self.identity_secret

                if self.save_password_item.get_active():
                    # Just for curious people. It's not even safe.
                    key = codecs.encode(self.__password.encode(), 'base64')
                    out = codecs.encode(key.decode(), 'rot13')
                    new_configs["password"] = out.replace('\n', '')

                if login_data.oauth:
                    new_configs['steamid'] = login_data.oauth['steamid']
                    new_configs['token'] = login_data.oauth['wgtoken']
                    new_configs['token_secure'] = login_data.oauth['wgtoken_secure']
                    new_configs['oauth_token'] = login_data.oauth['oauth_token']
                else:
                    new_configs['steamid'] = login_data.auth['transfer_parameters']['steamid']
                    new_configs['token'] = login_data.auth['transfer_parameters']['webcookie']
                    new_configs['token_secure'] = login_data.auth['transfer_parameters']['token_secure']

                for key_, value_ in new_configs.items():
                    config.new("login", key_, value_)

                steamid = universe.generate_steamid(new_configs['steamid'])
                _login_session.restore_login(steamid, new_configs['token'], new_configs['token_secure'])

                self.application.main_window.statusbar.clear('steamguard')

                self.has_user_data = True
                self.destroy()
            finally:
                self.save_password_item.set_sensitive(True)
                self.login_button.set_sensitive(True)
                self.set_size_request(400, 100)

            break

    def on_advanced_login_clicked(self, *args: Any) -> None:
        if self.advanced_login_section.props.visible:
            self.identity_secret_item.set_text('')
            self.shared_secret_item.set_text('')
            self.advanced_login_section.hide()
            self.set_size_request(400, 100)
        else:
            self.advanced_login_section.show()



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import logging
import time
from gi.repository import Gtk
from typing import Union, Optional, Any

from stlib import community
from stlib import universe
from . import utils, confirmation
from .. import config, i18n

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class CouponDialog(Gtk.Dialog):
    def __init__(
            self,
            parent_window: Gtk.Window,
            application: Gtk.Application,
            model: Optional[Gtk.TreeModel] = None,
            iter_: Union[Gtk.TreeIter, bool, None] = False,
    ) -> None:
        super().__init__(use_header_bar=True)
        self.parent_window = parent_window
        self.application = application
        self.community_session = community.Community.get_session(0)
        self.model = model
        self.iter = iter_
        self.has_status = False

        self.header_bar = self.get_header_bar()
        self.header_bar.set_show_title_buttons(False)

        self.set_default_size(300, 60)
        self.set_title(_('Get Coupon'))
        self.set_transient_for(self.parent_window)
        self.set_modal(True)
        self.set_destroy_with_parent(True)
        self.set_resizable(False)

        self.content_area = self.get_content_area()
        self.content_area.set_orientation(Gtk.Orientation.VERTICAL)
        self.content_area.set_spacing(10)
        self.content_area.set_margin_start(10)
        self.content_area.set_margin_end(10)
        self.content_area.set_margin_top(10)
        self.content_area.set_margin_bottom(10)

        self.status = utils.SimpleStatus()
        self.content_area.append(self.status)

        self.yes_button = Gtk.Button()
        self.yes_button.set_label(_("Continue"))
        self.header_bar.pack_end(self.yes_button)

        self.no_button = Gtk.Button()
        self.no_button.set_label(_("Cancel"))
        self.no_button.connect("clicked", lambda button: self.destroy())
        self.header_bar.pack_start(self.no_button)

        trade_time = int(time.time()) - config.parser.getint('coupons', 'last_trade_time')
        if self.iter and trade_time < 120:
            self.status.error(_("You must wait {} seconds to get another coupon").format(120 - trade_time))
            self.header_bar.set_show_title_buttons(True)
            self.yes_button.hide()
            self.no_button.hide()
            return

        if self.iter is None:
            self.status.error(_("You must select something"))
            self.header_bar.set_show_title_buttons(True)
            self.yes_button.hide()
            self.no_button.hide()
        elif self.iter is False:
            self.status.info(
                _(
                    "You are about to giveaway all coupons in your inventory\n"
                    "The items will be automatically accepted and transfered in 1 minute"
                )
            )
            self.yes_button.connect("clicked", self.on_yes_button_clicked, 'give')
        else:
            assert isinstance(self.model, Gtk.TreeModel)
            self.status.info(
                _(
                    "You are about to request {}\n"
                    "The item will be automatically added at your inventory in 1 minute"
                ).format(utils.unmarkup(self.model.get_value(self.iter, 1)))
            )
            self.yes_button.connect("clicked", self.on_yes_button_clicked, 'get')

        self.connect('response', lambda dialog, response_id: self.destroy())

    def on_yes_button_clicked(self, button: Gtk.Button, mode: str) -> None:
        button.hide()
        self.no_button.hide()

        self.set_size_request(0, 0)
        self.header_bar.set_show_title_buttons(False)

        loop = asyncio.get_event_loop()
        task = loop.create_task(self.send_trade_offer(mode))
        task.add_done_callback(self.on_task_finish)

    def on_task_finish(self, task: asyncio.Task[Any]) -> None:
        exception = task.exception()

        if exception and not isinstance(exception, asyncio.CancelledError):
            try:
                current_exception = task.exception()
                assert isinstance(current_exception, BaseException)
                raise current_exception
            except community.InventoryEmptyError:
                self.status.error(_("Inventory is empty."))
                self.header_bar.set_show_title_buttons(True)
            except Exception as exception:
                stack = task.get_stack()

                for frame in stack:
                    log.error(f"{type(exception).__name__} at {frame}")

                log.error(f"Steam Server is slow. {str(exception)}")

                self.status.error(
                    _(
                        "Unable to complete this trade. The reason is one of the following:\n\n"
                        "1. The item you choose already gone. Try another one.\n"
                        "2. You wrote a wrong token in config. Update you config.\n"
                        "3. The Steam server is slow. Wait a minute and try again.\n\n"
                        "If you keep seeing this error, please update the coupon list."
                    )
                )

                self.header_bar.set_show_title_buttons(True)
                self.yes_button.hide()
        else:
            config.new("coupons", "last_trade_time", int(time.time()))

            try:
                if self.iter:
                    assert isinstance(self.model, Gtk.TreeModel)
                    self.model.remove(self.iter)
            except IndexError:
                log.debug(_("Unable to remove tree path %s (already removed?). Ignoring."), self.iter)

            if not self.has_status:
                self.destroy()

    async def send_trade_offer(self, mode: str) -> None:
        self.status.info(_("Waiting Steam Server"))
        contextid = config.parser.getint('coupons', 'contextid')
        appid = config.parser.getint('coupons', 'appid')
        botid_raw = config.parser.get('coupons', 'botid_to_donate')
        token = config.parser.get('coupons', 'token_to_donate')
        steamid_raw = config.parser.getint('login', 'steamid')

        try:
            steamid = universe.generate_steamid(steamid_raw)
        except ValueError:
            self.status.info(_("Your steamid is invalid. (are you logged in?)"))
            self.header_bar.set_show_title_buttons(True)
            self.yes_button.hide()
            self.has_status = True
            return

        give = []
        receive = []

        if mode == 'get':
            assert isinstance(self.model, Gtk.TreeModel)
            botid_raw = self.model.get_value(self.iter, 3)
            token = self.model.get_value(self.iter, 4)
            assetid = int(self.model.get_value(self.iter, 5))
            receive = [(appid, assetid, 1)]
        else:
            json_data = await self.community_session.get_inventory(
                steamid,
                appid,
                contextid,
            )

            for coupon in json_data:
                give.append((coupon.appid, coupon.assetid, coupon.amount))

        try:
            botid = universe.generate_steamid(botid_raw)
        except ValueError:
            self.status.info(_("botid to donation is invalid. Check your config."))
            self.header_bar.set_show_title_buttons(True)
            self.yes_button.hide()
            self.has_status = True
            return

        json_data = await self.community_session.send_trade_offer(botid, token, contextid, give, receive)

        if len(json_data) == 1 and 'tradeofferid' in json_data:
            return

        if 'needs_email_confirmation' in json_data and json_data['needs_email_confirmation']:
            self.status.info(_('You will need to manually confirm the trade offer. Check your email.'))
            self.header_bar.set_show_title_buttons(True)
            self.yes_button.hide()

            # FIXME: track tradeoffer and wait for email confirmation
            self.has_status = True
            return

        if 'needs_mobile_confirmation' in json_data and json_data['needs_mobile_confirmation']:
            if not config.parser.getboolean('confirmations', 'enable'):
                self.status.info(_(
                    "Mobile confirmation is needed but the confirmation module isn't enabled.\n"
                    "You will need to manually confirm the trade offer."
                ))

                self.header_bar.set_show_title_buttons(True)
                self.yes_button.hide()
                # FIXME: track and wait for manual confirmation
                self.has_status = True
                return

            confirmation_store = self.parent_window.confirmations_tree.store
            target = None

            while True:
                confirmation_count = confirmation_store.iter_n_children()

                for index in range(confirmation_count):
                    iter_ = confirmation_store[index].iter
                    if confirmation_store.get_value(iter_, 1) == json_data['tradeofferid']:
                        target = iter_
                        break

                if target:
                    break
                else:
                    self.status.info(_('Waiting trade confirmation'))
                    await asyncio.sleep(5)

            finalize_dialog = confirmation.FinalizeDialog(
                self.parent_window,
                self.application,
                'allow',
                confirmation_store, target,
            )

            finalize_dialog.show()
            finalize_dialog.yes_button.emit('clicked')



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#

from importlib import resources

# Never use VHL methods in this file to avoid infinite recursion:
# [method>get_translation->vhlm->get_translation->vhlm] IT'S NOT A BUG!
import configparser
import gettext
import hashlib
from typing import Dict

from . import config

cache: Dict[bytes, str] = {}


def new_hash(text: str) -> bytes:
    sums = hashlib.sha256(text.encode())

    return sums.digest()


def get_translation(text: str) -> str:
    try:
        language = config.parser.get('general', 'language')
    except configparser.NoSectionError:
        # assume that config is not fully loaded yet
        return text

    with resources.as_file(resources.files('steam_tools_ng')) as path:
        translation = gettext.translation("steam-tools-ng", path / 'locale', languages=[language], fallback=True)
        translated_text = translation.gettext(text)
        cache[new_hash(translated_text)] = text

    return translated_text



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import configparser
import locale
import logging
import os
import site
from collections import OrderedDict
from pathlib import Path

import sys
from typing import Any, Dict

from stlib import plugins as stlib_plugins
from . import i18n, logger_handlers

parser = configparser.RawConfigParser()
log = logging.getLogger(__name__)

if Path('src').is_dir():
    # development mode
    data_dir = Path('config')
elif hasattr(sys, 'frozen') or sys.platform == 'win32':
    data_dir = Path(os.environ['LOCALAPPDATA'])
else:
    data_dir = Path(os.getenv('XDG_CONFIG_HOME', Path.home() / '.config'))

config_file_directory = data_dir / 'steam-tools-ng'
config_file_name = 'steam-tools-ng.config'
config_file = config_file_directory / config_file_name

try:
    from stlib import client
except ImportError as exception:
    log.error(str(exception))
    client = None


def _(message):
    return message


gtk_themes = OrderedDict([
    ('light', _("Light")),
    ('dark', _("Dark")),
])

log_levels = OrderedDict([
    ('critical', _("Critical")),
    ('error', _("Error")),
    ('warning', _("Warning")),
    ('info', _("Info")),
    ('debug', _("Debug")),
])

translations = OrderedDict([
    ('en', _("English")),
    ('pt_BR', _("Portuguese (Brazil)")),
    ('fr', _("French")),
])

giveaway_types = OrderedDict([
    ('main', _("No restriction")),
    ('new', _("New Giveaways")),
    ('recommended', _("Recommended")),
    ('wishlist', _("Wishlist")),
    ('group', _('Group')),
    ('dlc', _('DLC')),
    ('region_restricted', _("Region Restricted")),
])

giveaway_sort_types = OrderedDict([
    ('name+', _("Name (A-Z)")),
    ('name-', _("Name (Z-A)")),
    ('copies+', _("Copies (0..99)")),
    ('copies-', _("Copies (99..0)")),
    ('points+', _("Points (0-50)")),
    ('points-', _("Points (50-0)")),
    ('level+', _("Level (0-100)")),
    ('level-', _("Level (100-0)")),
])

steamgifts_modes = OrderedDict([
    ('run_all_and_restart', _('Run all and restart')),
    ('stop_after_minimum_and_wait', _('Stop after reach minimum points and wait')),
    ('stop_after_minimum_and_restart', _('Stop after reach minimum points and restart')),
])

plugins = OrderedDict([
    ("coupons", _("Free Coupons")),
    ("confirmations", _("Confirmations")),
    ("steamtrades", _("Steam Trades")),
    ("steamgifts", _("Steam Gifts")),
    ("steamguard", _("Steam Guard")),
    ("cardfarming", _("Card Farming")),
])

coupon_discounts = OrderedDict([
    ("33", _("33% OFF")),
    ("50", _("50% OFF")),
    ("66", _("66% OFF")),
    ("75", _("75% OFF")),
    ("90", _("90% OFF")),
])

_ = i18n.get_translation

if sys.platform == 'win32':
    event_loop = asyncio.ProactorEventLoop()
    file_manager = 'explorer'
else:
    file_manager = 'xdg-open'
    event_loop = asyncio.new_event_loop()

asyncio.set_event_loop(event_loop)

default_config: Dict[str, Dict[str, Any]] = {
    'logger': {
        'log_directory': data_dir / 'steam-tools-ng',
        'log_level': 'debug',
        'log_console_level': 'info',
        'log_color': True,
    },
    'steam': {
        'api_url': 'https://api.steampowered.com',
    },
    'coupons': {
        'enable': True,
        'botid_to_donate': '76561198018370992',
        'botids': '76561198018370992',
        'appid': '753',
        'contextid': '3',
        'token_to_donate': '6Z6Xn5NM',
        'tokens': '6Z6Xn5NM',
        'blacklist': '',
        'last_trade_time': 0,
        'minimum_discount': 75,
    },
    'steamguard': {
        'enable': True,
        'enable_confirmations': True,
    },
    'steamtrades': {
        'enable': True,
        'wait_for_bump': 3700,
        'trade_ids': '',
    },
    'steamgifts': {
        'enable': True,
        'developer_giveaways': 'True',
        'mode': 'run_all_and_restart',
        'wait_after_each_strategy': 10,
        'wait_after_full_cycle': 3700,
        'minimum_points': 0,
    },
    'cardfarming': {
        'enable': True,
        'reverse_sorting': False,
        'mandatory_waiting': 7200,
        'wait_while_running': 300,
        'wait_for_drops': 120,
        'max_concurrency': 50,
        'invisible': True,
    },
    'fakerun': {
        'cakes': '',
    },
    'general': {
        'theme': 'light',
        'show_close_button': True,
        'language': str(locale.getdefaultlocale()[0]),
    },
    'login': {
        'steamid': 0,
        'deviceid': '',
        'token': '',
        'token_secure': '',
        'oauth_token': '',
        'account_name': '',
        'shared_secret': '',
        'identity_secret': '',
        'password': '',
    },
}

for index in range(1, 6):
    default_config[f'steamgifts_strategy{index}'] = {
        'enable': False,
        'minimum_points': 0,
        'maximum_points': 50,
        'minimum_level': 0,
        'maximum_level': 100,
        'minimum_copies': 1,
        'maximum_copies': 999999,
        'minimum_metascore': 0,
        'maximum_metascore': 100,
        'minimum_entries': 0,
        'maximum_entries': 999999,
        'restrict_type': "main",
        'sort_type': "name+",
    }

    if index == 1:
        default_config[f'steamgifts_strategy{index}']['enable'] = True


def update_log_level(type_: str, level_string: str) -> None:
    level = getattr(logging, level_string.upper())
    file_handler, console_handler, *extra_handlers = logging.root.handlers

    if type_ == "console":
        console_handler.setLevel(level)
    else:
        file_handler.setLevel(level)


def validate_config(section: str, option: str, defaults: OrderedDict[str, str]) -> None:
    value = parser.get(section, option)

    if value and value not in defaults.keys():
        if option == 'language':
            log.error(_("Unsupported language requested. Fallbacking to English."))
            new('general', 'language', 'en')
            return

        raise configparser.Error(_("Please, fix your config file. Available values for {}:\n{}").format(
            option,
            ', '.join(defaults.keys()),
        ))


def init() -> None:
    config_file_directory.mkdir(parents=True, exist_ok=True)
    parser.read_dict(default_config)

    if config_file.is_file():
        parser.read(config_file)

    # fallback deprecated values
    if (parser.get('steam', 'api_url') in [
        'https://api.lara.monster', 'https://api.lara.click',
    ]):
        new('steam', 'api_url', default_config['steam']['api_url'])

    log_directory = Path(parser.get("logger", "log_directory"))

    if not log_directory.is_dir():
        log.error(_("Incorrect log directory. Fallbacking to default."))
        log_directory = data_dir / 'steam-tools-ng'
        new("logger", "log_directory", log_directory)

    validate_config("logger", "log_level", log_levels)
    validate_config("logger", "log_console_level", log_levels)
    validate_config("general", "theme", gtk_themes)
    validate_config("general", "language", translations)
    validate_config("steamgifts", "mode", steamgifts_modes)

    for _index in range(1, 4):
        validate_config(f"steamgifts_strategy{_index}", "restrict_type", giveaway_types)
        validate_config(f"steamgifts_strategy{_index}", "sort_type", giveaway_sort_types)

    log_directory.mkdir(parents=True, exist_ok=True)

    stlib_plugins.add_search_paths(
        str(Path(os.getcwd(), 'lib', 'stlib-plugins')),
        *[str(Path(site_, 'stlib-plugins')) for site_ in site.getsitepackages()],
        str(Path(site.getusersitepackages(), 'stlib-plugins')),
    )

    if not stlib_plugins.has_plugin("steamtrades"):
        new("steamtrades", "enable", False)

    if not stlib_plugins.has_plugin("steamgifts"):
        new("steamgifts", "enable", False)

    if not client:
        new("cardfarming", "enable", False)


def init_logger() -> None:
    log_directory = Path(parser.get("logger", "log_directory"))
    log_level = parser.get("logger", "log_level")
    log_console_level = parser.get("logger", "log_console_level")

    log_file_handler = logger_handlers.RotatingFileHandler(log_directory / 'steam-tools-ng.log',
                                                           backupCount=1,
                                                           encoding='utf-8')
    log_file_handler.setFormatter(logging.Formatter('%(name)s:%(levelname)s (%(funcName)s) => %(message)s'))
    log_file_handler.setLevel(getattr(logging, log_level.upper()))

    try:
        log_file_handler.doRollover()
    except PermissionError:
        log.debug(_("Unable to open steam-tools-ng.log"))
        log_file_handler.close()
        log_file_handler = logger_handlers.NullHandler()  # type: ignore

    log_console_handler = logger_handlers.ColoredStreamHandler()
    log_console_handler.setLevel(getattr(logging, log_console_level.upper()))

    if 'gtk' not in sys.modules:
        log_console_handler.setLevel(logging.WARNING)

    # noinspection PyArgumentList
    logging.basicConfig(level=logging.DEBUG, handlers=[log_file_handler, log_console_handler])


def new(section: str, option: str, value: Any) -> None:
    if option == "log_level":
        update_log_level("file", value)
    elif option == "log_console_level":
        update_log_level("console", value)

    if parser.get(section, option, fallback='') != str(value):
        log.debug(_('Saving {}:{} on config file').format(section, option))
        parser.set(section, option, str(value))

        with open(config_file, 'w', encoding="utf8") as config_file_object:
            parser.write(config_file_object)
    else:
        log.debug(_('Not saving {}:{} because values are already updated').format(section, option))


def remove(section: str, option: str) -> None:
    # Some GUI checks will fail if option doesn't exists
    new(section, option, '')
    # parser.remove_option(section, option)

    # with open(config_file, 'w', encoding="utf8") as config_file_object:
    #    parser.write(config_file_object)



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#

import asyncio
import contextlib
from gi.repository import Gtk, GLib
from typing import Optional


async def async_iterator(
        application: Gtk.Application,
        main_context: GLib.MainContext,
        loop: asyncio.AbstractEventLoop,
) -> None:
    while main_context.pending():
        main_context.iteration(False)

    await asyncio.sleep(0.01)

    if application.main_window and application.main_window.get_realized():
        loop.create_task(async_iterator(application, main_context, loop))
    else:
        application.quit()
        loop.stop()


async def async_iterator_for_the_fifth_dimension(
        main_context: GLib.MainContext,
        loop: asyncio.AbstractEventLoop,
) -> None:
    while main_context.pending():
        main_context.iteration(False)

    await asyncio.sleep(0.01)
    loop.create_task(async_iterator_for_the_fifth_dimension(main_context, loop))


# FIXME: https://github.com/python/asyncio/pull/465
def run(application: Optional[Gtk.Application] = None) -> None:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    main_context = GLib.MainContext.default()

    if not application:
        iterator = async_iterator_for_the_fifth_dimension(main_context, loop)
    else:
        iterator = async_iterator(application, main_context, loop)
        application.register()
        application.activate()

    loop.create_task(iterator)

    with contextlib.suppress(KeyboardInterrupt):
        loop.run_forever()



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#

import asyncio
import inspect
import logging
import time
from dataclasses import dataclass
from functools import cache, wraps
from typing import Tuple, Any, Callable, AsyncGenerator


@dataclass
class ModuleData:
    display: str = ''
    status: str = ''
    info: str = ''
    error: str = ''
    level: Tuple[int, int] = (0, 0)
    action: str = ''
    raw_data: Any = None
    suppress_logging: bool = False


async def timed_module_data(wait_offset: int, module_data: ModuleData) -> AsyncGenerator[ModuleData, None]:
    info = module_data.info
    assert module_data.level == (0, 0), "level should not be used here"

    # Prevent action to being executed multiple times
    if module_data.action:
        yield module_data
        module_data.action = ''

    module_data.suppress_logging = True
    caller = inspect.currentframe().f_back
    log = logging.getLogger(caller.f_globals['__name__'])
    log.info(info)

    for past_time in range(wait_offset):
        current_time = round((wait_offset - past_time) / 60)
        current_time_size = 'm'

        if current_time <= 1:
            current_time = wait_offset - past_time
            current_time_size = 's'

        module_data.level = (past_time, wait_offset)
        module_data.info = f'{info} ({current_time}{current_time_size})'

        yield module_data
        await asyncio.sleep(1)


def time_offset_cache(ttl: int = 60) -> Callable[[Callable[[], int]], Callable[[], int]]:
    def wrapper(function_: Any) -> Callable[[], int]:
        function_ = cache(function_)
        function_.time_base = time.time()

        @wraps(function_)
        def wrapped() -> int:
            if time.time() >= function_.time_base + ttl:
                function_.cache_clear()
                function_.time_base = time.time()

            time_raw = function_()

            if function_.cache_info().currsize == 0:
                assert isinstance(time_raw, int)
                return time_raw

            function_.time_offset = function_.time_base - time_raw
            return round(time.time() + function_.time_offset)

        return wrapped

    return wrapper



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import contextlib
import functools
import logging
import sys
from typing import Optional, Any, Callable

import aiohttp

import stlib
from stlib import plugins, universe, login, community, webapi, internals
from . import authenticator, utils
from . import login as cli_login
from .. import i18n, config, core

log = logging.getLogger(__name__)
_ = i18n.get_translation


def while_running(function: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(function)
    async def wrapper(self: 'SteamToolsNG', *args: Any, **kwargs: Any) -> None:
        while True:
            await function(self, *args, **kwargs)

            if self.stop:
                self.on_quit()
                break

    return wrapper


# noinspection PyUnusedLocal
class SteamToolsNG:
    def __init__(self, module_name: str, module_options: str) -> None:
        self.module_name = module_name
        self.stop = False
        self.custom_gameid = 0
        self.extra_gameid = None

        if module_name in ['cardfarming', 'fakerun'] and not stlib.steamworks_available:
            log.critical(_(
                "{} module has been disabled because you have "
                "a stlib built without SteamWorks support. To enable it again, "
                "reinstall stlib with SteamWorks support"
            ).format(module_name))
            sys.exit(1)

        if module_name in ['steamtrades', 'steamgifts']:
            if not plugins.has_plugin(module_name):
                log.critical(_(
                    "{0} module has been disabled because you don't "
                    "have {0} plugin installed. To enable it again, "
                    "install the {0} plugin."
                ).format(module_name))
                sys.exit(1)

        try:
            if module_name == 'fakerun' and not module_options:
                raise ValueError

            for index, option in enumerate(module_options):
                self.stop = True

                if option != 'oneshot':
                    self.custom_gameid = int(option)

                    if module_name == 'fakerun':
                        if self.custom_gameid == 34:
                            if len(module_options) < 2:
                                raise ValueError

                            self.extra_gameid = int(module_options[index + 1])

                            break
        except ValueError:
            logging.critical("Wrong command line params!")
            sys.exit(1)

        self.api_url = config.parser.get("steam", "api_url")

    @property
    def steamid(self) -> Optional[universe.SteamId]:
        steamid = config.parser.getint("login", "steamid")

        if steamid:
            try:
                return universe.generate_steamid(steamid)
            except ValueError:
                log.warning(_("SteamId is invalid"))

        return None

    # FIXME: https://github.com/python/asyncio/pull/465
    def run(self) -> None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        task = loop.create_task(self.async_activate())
        task.add_done_callback(utils.safe_task_callback)

        with contextlib.suppress(KeyboardInterrupt):
            loop.run_forever()

    async def do_login(self, *, block: bool = True, auto: bool = False) -> None:
        login_session = cli_login.Login(self)
        await login_session.do_login(auto)

    async def async_activate(self) -> None:
        login_session = await login.Login.new_session(0, api_url=self.api_url)

        utils.set_console(info=_("Logging on Steam. Please wait!"))

        token = config.parser.get("login", "token")
        token_secure = config.parser.get("login", "token_secure")

        if not token or not token_secure or not self.steamid:
            await self.do_login()

        login_session.restore_login(self.steamid, token, token_secure)

        try:
            if await login_session.is_logged_in(self.steamid):
                utils.set_console(info=_("Steam login Successful"))
            else:
                await self.do_login(auto=True)
        except aiohttp.ClientError as exception:
            log.exception(str(exception))
            log.error(_("Check your connection. (server down?)"))
            await asyncio.sleep(10)
            return  # FIXME: RETRY ###

        community_session = await community.Community.new_session(0, api_url=self.api_url)

        try:
            api_key = await community_session.get_api_key()
            log.debug(_('SteamAPI key found: %s'), api_key)

            if api_key[1] != 'Steam Tools NG':
                raise AttributeError
        except AttributeError:
            log.warning(_('Updating your SteamAPI dev key'))
            await asyncio.sleep(3)
            await community_session.revoke_api_key()
            await asyncio.sleep(3)
            api_key = await community_session.register_api_key('Steam Tools NG')

            if not api_key:
                raise ValueError(_('Something wrong with your SteamAPI dev key'))

        await webapi.SteamWebAPI.new_session(0, api_key=api_key[0], api_url=self.api_url)
        await internals.Internals.new_session(0)

        if self.module_name in ['steamtrades', 'steamgifts']:
            plugin = plugins.get_plugin(self.module_name)
            await plugin.Main.new_session(0)

        log.debug(_("Initializing module %s"), self.module_name)
        module = getattr(self, f"run_{self.module_name}")
        task = asyncio.create_task(module())
        log.debug(_("Adding a new callback for %s"), task)
        task.add_done_callback(utils.safe_task_callback)

    async def run_add_authenticator(self) -> None:
        authenticator_config = authenticator.NewAuthenticator(self)
        await authenticator_config.add_authenticator()

    @while_running
    async def run_steamguard(self) -> None:
        steamguard = core.steamguard.main()

        async for module_data in steamguard:
            utils.set_console(module_data)

    @while_running
    async def run_cardfarming(self) -> None:
        cardfarming = core.cardfarming.main(self.steamid, self.custom_gameid)

        async for module_data in cardfarming:
            utils.set_console(module_data)

    @while_running
    async def run_fakerun(self) -> None:
        fakerun = core.fakerun.main(self.steamid, self.custom_gameid, self.extra_gameid)

        async for module_data in fakerun:
            utils.set_console(module_data)

    @while_running
    async def run_steamtrades(self) -> None:
        steamtrades = core.steamtrades.main()

        async for module_data in steamtrades:
            utils.set_console(module_data)

            if module_data.action == "login":
                await self.do_login(auto=True)
                continue

    @while_running
    async def run_steamgifts(self) -> None:
        steamgifts = core.steamgifts.main()

        async for module_data in steamgifts:
            utils.set_console(module_data)

            if module_data.action == "login":
                await self.do_login(auto=True)
                continue

    @staticmethod
    def on_quit(exit_code: int = 0) -> None:
        loop = asyncio.get_running_loop()
        loop.stop()
        # sys.exit(exit_code)



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import aiohttp
import asyncio
import atexit
import contextlib
import random
import sys
import time
from pathlib import Path
from subprocess import call
from typing import AsyncGenerator, Dict, Optional, Any

from stlib import webapi, client, universe, community
from . import utils
from .. import i18n, config

if sys.platform == 'win32':
    from psutil import Popen, NoSuchProcess

_ = i18n.get_translation


# FIXME: Workaround for keyboard lag when running games on GUI on Windows
# FIXME: https://gitlab.gnome.org/GNOME/gtk/-/issues/2015
# FIXME: L36:91
def killall(process: 'Popen') -> None:
    if not process.is_running():
        return

    for child in process.children(recursive=True):
        with contextlib.suppress(NoSuchProcess):
            child.kill()
            child.wait()

    process.kill()
    process.wait()


class SteamAPIExecutorWorkaround:
    def __init__(self, appid: int, *args: Any, **kwargs: Any) -> None:
        self.process = None
        self.appid = appid
        self._is_running = False

        if sys.platform == 'win32':
            if getattr(sys, 'frozen', False):
                executor_path = [str(Path(sys.executable).parent / 'steam-api-executor.exe'), str(self.appid)]
            else:
                executor_path = [sys.executable, '-m', 'steam_tools_ng.steam_api_executor', str(self.appid)]

            self.process = Popen(executor_path, creationflags=0x08000000)
            self._is_running = True
            atexit.register(killall, self.process)
        else:
            self.executor = _SteamAPIExecutor(appid, *args, **kwargs)

    def is_running(self) -> bool:
        if sys.platform == 'win32':
            return self._is_running
        else:
            return self.executor.is_running()

    def shutdown(self, *args: Any, **kwargs: Any) -> None:
        if sys.platform == 'win32':
            killall(self.process)
            self._is_running = False
        else:
            self.executor.shutdown(*args, **kwargs)


if 'gtk' in sys.modules:
    _SteamAPIExecutor = client.SteamAPIExecutor
    client.SteamAPIExecutor = SteamAPIExecutorWorkaround

# FIXME: L36:91
# FIXME: ---- #

executors = {}
total_cards_remaining = 0


def safe_exit(*args: Any, **kwargs: Any) -> None:
    for executor in executors.values():
        executor.shutdown(*args, **kwargs)


async def while_has_cards(
        steamid: universe.SteamId,
        badge: community.Badge,
) -> AsyncGenerator[utils.ModuleData, None]:
    webapi_session = webapi.SteamWebAPI.get_session(0)
    community_session = community.Community.get_session(0)

    while badge.cards != 0:
        mandatory_waiting = config.parser.getint("cardfarming", "mandatory_waiting")
        wait_while_running = config.parser.getint("cardfarming", "wait_while_running")
        wait_for_drops = config.parser.getint("cardfarming", "wait_for_drops")

        try:
            game_list = await webapi_session.get_owned_games(steamid, appids_filter=[badge.appid])
            game_info = game_list[0]
        except aiohttp.ClientError:
            module_data = utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))

            async for data in utils.timed_module_data(10, module_data):
                yield data

            continue

        if game_info.playtime_forever * 60 >= mandatory_waiting:
            wait_offset = random.randint(wait_while_running, int(wait_while_running / 100 * 125))
        else:
            wait_offset = mandatory_waiting - game_info.playtime_forever * 60

        try:
            executor = client.SteamAPIExecutor(badge.appid)
        except AttributeError:
            yield utils.ModuleData(action='ignore', info=_("Invalid game id {}. Ignoring.").format(badge.appid))
            break
        except ProcessLookupError:
            module_data = utils.ModuleData(error=_("Steam Client is not running."), info=_("Waiting Changes"))

            async for data in utils.timed_module_data(15, module_data):
                yield data

            continue

        module_data = utils.ModuleData(
            display=str(badge.appid),
            info=badge.name,
            status=_("Running {}").format(badge.name),
            raw_data=executor,
            action="check",
        )

        async for data in utils.timed_module_data(wait_offset, module_data):
            yield data

        executor.shutdown()
        wait_offset = random.randint(wait_for_drops, int(wait_for_drops / 100 * 125))

        module_data = utils.ModuleData(
            display=str(badge.appid),
            info=_('Updating {} drops').format(badge.name),
            status=_("Game paused"),
        )

        async for data in utils.timed_module_data(wait_offset, module_data):
            yield data

        while True:
            try:
                cards = await community_session.get_card_drops_remaining(steamid, badge.appid)
            except aiohttp.ClientError:
                yield utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))
                await asyncio.sleep(10)
            except community.BadgeError:
                yield utils.ModuleData(error=_("Steam Server is busy"), info=_("Waiting Changes"))
                await asyncio.sleep(20)
            else:
                break

        global total_cards_remaining
        total_cards_remaining -= badge.cards - cards
        # noinspection PyProtectedMember
        badge = badge._replace(cards=cards)

    utils.ModuleData(
        display=str(badge.appid),
        info=_("{} ({})").format(_("Done"), badge.name),
    )


async def main(steamid: universe.SteamId, custom_game_id: int = 0) -> AsyncGenerator[utils.ModuleData, None]:
    asyncio.current_task().add_done_callback(safe_exit)

    reverse_sorting = config.parser.getboolean("cardfarming", "reverse_sorting")
    max_concurrency = config.parser.getint("cardfarming", "max_concurrency")
    invisible = config.parser.getboolean("cardfarming", "invisible")
    community_session = community.Community.get_session(0)
    global total_cards_remaining

    try:
        badges = sorted(
            await community_session.get_badges(steamid),
            key=lambda badge_: badge_.cards,  # type: ignore
            reverse=reverse_sorting
        )
    except aiohttp.ClientError:
        module_data = utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))

        async for data in utils.timed_module_data(10, module_data):
            yield data

        return

    if not badges or (custom_game_id and custom_game_id not in [badge.appid for badge in badges]):
        module_data = utils.ModuleData(error=_("No more cards to drop."), info=_("Waiting Changes"))
        wait_offset = random.randint(300, 500)

        async for data in utils.timed_module_data(wait_offset, module_data):
            yield data

        return

    generators = {}

    if invisible:
        call(f'{config.file_manager} "steam://friends/status/invisible"')

    for badge in badges:
        yield utils.ModuleData(
            display=str(badge.appid),
            status=_("Loading {}").format(badge.name),
        )

        if custom_game_id and badge.appid != custom_game_id:
            yield utils.ModuleData(info=_("Skipping {}").format(badge.appid))
            continue

        generators[badge.appid] = while_has_cards(steamid, badge)
        total_cards_remaining += badge.cards

    tasks: Dict[int, Optional[asyncio.Task[Any]]] = {}
    semaphore = asyncio.Semaphore(max_concurrency)
    last_update = 0

    while True:
        for appid in generators.keys():
            progress_coro = anext(generators[appid])
            assert asyncio.iscoroutine(progress_coro)

            if appid not in tasks:
                if semaphore.locked():
                    break

                await semaphore.acquire()
                tasks[appid] = asyncio.create_task(progress_coro)

            if not tasks[appid]:
                continue

            current_task = tasks[appid]
            assert isinstance(current_task, asyncio.Task)

            if current_task.done():
                semaphore.release()

                if current_task.exception():
                    if isinstance(current_task.exception(), StopAsyncIteration):
                        tasks[appid] = None
                        continue

                    current_exception = current_task.exception()
                    assert isinstance(current_exception, BaseException)
                    raise current_exception

                await semaphore.acquire()
                tasks[appid] = asyncio.create_task(progress_coro)

        if not any(tasks.values()):
            break

        await asyncio.wait([task for task in tasks.values() if task], return_when=asyncio.FIRST_COMPLETED)

        for appid, task in tasks.items():
            if task and task.done() and not task.exception():
                global executors
                data: utils.ModuleData = task.result()

                if data.action == 'check':
                    executors[appid] = data.raw_data

                if int(time.time()) > last_update + 3:
                    current_running_limit = len(tasks)
                    total_remaining = len(generators) - len([task for task in tasks.values() if not task])
                    running_executors = [executor for executor in executors.values() if executor.is_running()]
                    extra_info = ''

                    if current_running_limit > total_remaining:
                        current_running_limit = total_remaining

                    if current_running_limit == 2:
                        extra_info = _(" +{} other").format(current_running_limit - 1)
                    elif current_running_limit > 2:
                        extra_info = _(" +{} others").format(current_running_limit - 1)

                    yield utils.ModuleData(
                        display=' : '.join([str(executor.appid) for executor in running_executors]),
                        info=data.info + extra_info,
                        status=_('{} from {} remaining ({} cards)').format(
                            current_running_limit,
                            total_remaining,
                            total_cards_remaining,
                        ),
                        level=data.level,
                        raw_data=running_executors,
                        action=data.action,
                    )
                    last_update = int(time.time())



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import aiohttp
import asyncio
import logging
import random
from typing import AsyncGenerator

from stlib import plugins, login
from . import utils
from .. import i18n, config

_ = i18n.get_translation
log = logging.getLogger(__name__)


async def main() -> AsyncGenerator[utils.ModuleData, None]:
    yield utils.ModuleData(status=_("Loading"))

    if plugins.has_plugin("steamtrades"):
        steamtrades = plugins.get_plugin("steamtrades")
        steamtrades_session = steamtrades.Main.get_session(0)
    else:
        raise ImportError(_("Unable to find Steamtrades plugin"))

    trade_ids = config.parser.get("steamtrades", "trade_ids")
    wait_for_bump = config.parser.getint("steamtrades", "wait_for_bump")

    if not trade_ids:
        yield utils.ModuleData(error=_("No trade ID found"), info=_("Waiting Changes"))
        await asyncio.sleep(5)
        return

    trades = [trade.strip() for trade in trade_ids.split(',')]

    try:
        await steamtrades_session.do_login()
    except aiohttp.ClientError:
        yield utils.ModuleData(error=_("Check your connection. (server down?)"))
        await asyncio.sleep(15)
        return
    except steamtrades.TooFast:
        yield utils.ModuleData(error=_("Unable to login. Trying again in 15 seconds"))
        await asyncio.sleep(15)
        return
    except steamtrades.UserSuspended:
        module_data = utils.ModuleData(error=_("User is suspended."))

        async for data in utils.timed_module_data(18000, module_data):
            yield data

        return
    except steamtrades.PrivateProfile:
        yield utils.ModuleData(error=_("Your profile must be public to use steamtrades."), info=_("Waiting Changes"))
        await asyncio.sleep(30)
        return
    except steamtrades.UserLevelError:
        yield utils.ModuleData(error=_("You must be level 1 or greater to use steamtrades."))
        await asyncio.sleep(30)
        return
    except login.LoginError:
        yield utils.ModuleData(error=_("User is not logged in. Trying again in 30 seconds"))
        await asyncio.sleep(30)
        return

    bumped = False

    for trade_id in trades:
        try:
            trade_info = await steamtrades_session.get_trade_info(trade_id)
        except (IndexError, aiohttp.ClientResponseError):
            yield utils.ModuleData(error=_("Unable to find trade id"))
            bumped = False
            break
        except aiohttp.ClientError:
            yield utils.ModuleData(error=_("Check your connection. (server down?)"))
            bumped = False
            break

        module_data = utils.ModuleData(display=trade_info.id, info=trade_info.title)
        max_ban_wait = random.randint(5, 15)

        async for data in utils.timed_module_data(max_ban_wait, module_data):
            yield data

        try:
            if await steamtrades_session.bump(trade_info):
                yield utils.ModuleData(display=trade_id, info=_("Bumped!"))
                bumped = True
            else:
                yield utils.ModuleData(display=trade_id, error=_("Unable to bump"))
                await asyncio.sleep(5)
                continue
        except aiohttp.ClientError:
            yield utils.ModuleData(error=_("Check your connection. (server down?)"))
            await asyncio.sleep(10)
            bumped = False
            break
        except steamtrades.NoTradesError:
            yield utils.ModuleData(error=_("No trades available to bump"))
            await asyncio.sleep(15)
            continue
        except steamtrades.TradeNotReadyError as exception:
            wait_for_bump = exception.time_left * 60
            bumped = True
        except steamtrades.TradeClosedError as exception:
            yield utils.ModuleData(error=_("Trade {}({}) is closed").format(exception.title, exception.id))
            await asyncio.sleep(5)
            continue
        except login.LoginError:
            yield utils.ModuleData(error=_("Login is lost. Trying to relogin."))
            await asyncio.sleep(5)
            bumped = False
            break

    if not bumped:
        await asyncio.sleep(10)
        return

    wait_offset = random.randint(wait_for_bump, wait_for_bump + 400)
    module_data = utils.ModuleData(info=_("Waiting Changes"))

    async for data in utils.timed_module_data(wait_offset, module_data):
        yield data



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import aiohttp
import asyncio
import logging
from typing import AsyncGenerator

from stlib import universe, community, internals, webapi
from . import utils
from .. import i18n, config

_ = i18n.get_translation
log = logging.getLogger(__name__)


async def main(steamid: universe.SteamId, fetch_coupon_event: asyncio.Event) -> AsyncGenerator[utils.ModuleData, None]:
    await fetch_coupon_event.wait()

    community_session = community.Community.get_session(0)
    internals_session = internals.Internals.get_session(0)
    webapi_session = webapi.SteamWebAPI.get_session(0)
    botids = config.parser.get('coupons', 'botids')
    tokens = config.parser.get('coupons', 'tokens')
    appid = config.parser.getint('coupons', 'appid')
    contextid = config.parser.getint('coupons', 'contextid')

    if not botids:
        yield utils.ModuleData(error=_("No botID found"), info=_("Waiting Changes"))
        await asyncio.sleep(5)
        return

    bot_list = [bot.strip() for bot in botids.split(',')]
    token_list = [token.strip() for token in tokens.split(',')]

    if len(bot_list) != len(token_list):
        yield utils.ModuleData(error=_("Invalid config. Each bot must have id and token."), info=_("Waiting Changes"))
        await asyncio.sleep(5)
        return

    try:
        owned_games = await webapi_session.get_owned_games(steamid)
    except aiohttp.ClientError:
        yield utils.ModuleData(error=_("Failed when trying to get owned games"))
        await asyncio.sleep(15)
        return

    yield utils.ModuleData(action="clear")
    package_count = 1

    for botid, token in zip(bot_list, token_list):
        try:
            steamid = universe.generate_steamid(botid)
        except ValueError:
            yield utils.ModuleData(error=_("The botid {} is invalid").format(botid))
            await asyncio.sleep(5)
            return

        try:
            inventory = await community_session.get_inventory(steamid, appid, contextid)
        except AttributeError:
            module_data = utils.ModuleData(error=_("Error when fetch inventory"), info=_("Waiting Changes"))

            async for data in utils.timed_module_data(15, module_data):
                yield data

            return
        except aiohttp.ClientError:
            module_data = utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))

            async for data in utils.timed_module_data(60, module_data):
                yield data

            return

        if not inventory:
            module_data = utils.ModuleData(error=_("The botid {} has no coupons available"), info=_("Skipping"))

            async for data in utils.timed_module_data(15, module_data):
                yield data

            continue

        for index, coupon_ in enumerate(inventory):
            yield utils.ModuleData(action="update_level", raw_data=(index, len(inventory)))
            package_link = coupon_.actions[0]['link']
            packageids = [int(id_) for id_ in package_link.split('=')[1].split(',')]
            blacklist = config.parser.get('coupons', 'blacklist')
            ignored_list = [coupon.split('% OFF')[-1].split('- Coupon')[0].strip() for coupon in blacklist.split(',')]
            ignored_list += [game.name for game in owned_games]
            minimum_discount = config.parser.getint('coupons', 'minimum_discount')

            for package_id in packageids:
                if not fetch_coupon_event.is_set():
                    log.warning(_("Stopping fetching coupons (requested by user)"))
                    yield utils.ModuleData(action="update_level", raw_data=(0, 0))
                    return

                if coupon_.name.split('% OFF')[-1].split('- Coupon')[0].strip() in ignored_list:
                    log.info(_('Ignoring coupon %s due blacklist'), coupon_.name)
                    continue

                coupon_discount = int(coupon_.name.split('%')[0])

                if coupon_discount < minimum_discount:
                    log.info(_('Ignoring coupon %s due low discount value'), coupon_.name)
                    continue

                package_count += 1

                try:
                    package_details = await internals_session.get_package(package_id)

                    if not package_details:
                        raise ValueError
                except aiohttp.ClientError:
                    module_data = utils.ModuleData(
                        error=_("Check your connection. (server down?)"),
                        info=_("Waiting Changes"),
                    )

                    async for data in utils.timed_module_data(60, module_data):
                        yield data

                    continue
                except ValueError:
                    yield utils.ModuleData(error=_("Failed to get package details"), info=_("Waiting Changes"))
                    await asyncio.sleep(1)
                    continue
                else:
                    await asyncio.sleep(.5)

                if package_details.discount_percent:
                    real_price = package_details.price - (
                            package_details.price * package_details.discount_percent / 100)
                else:
                    real_price = package_details.price - (package_details.price * coupon_discount / 100)

                yield utils.ModuleData(action='update', raw_data={
                    'price': round(real_price, 2),
                    'name': coupon_.name,
                    'link': coupon_.actions[0]['link'],
                    'botid': botid,
                    'token': token,
                    'assetid': coupon_.assetid,
                })

            if index and not package_count % 130:
                module_data = utils.ModuleData(error=_("Api rate limit reached. Waiting."), info=_("Waiting Changes"))

                async for data in utils.timed_module_data(120, module_data):
                    yield data

    fetch_coupon_event.clear()



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import asyncio
import logging
import random
from typing import AsyncGenerator

import aiohttp

from stlib import plugins, login
from . import utils
from .. import i18n, config

_ = i18n.get_translation
log = logging.getLogger(__name__)


async def main() -> AsyncGenerator[utils.ModuleData, None]:
    yield utils.ModuleData(status=_("Loading"))

    if plugins.has_plugin("steamgifts"):
        steamgifts = plugins.get_plugin("steamgifts")
        steamgifts_session = steamgifts.Main.get_session(0)
    else:
        raise ImportError(_("Unable to find Steamgifts plugin."))

    try:
        await steamgifts_session.do_login()
    except aiohttp.ClientError:
        yield utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))
        await asyncio.sleep(15)
        return
    except steamgifts.TooFast:
        yield utils.ModuleData(error=_("Unable to login. Trying again in 15 seconds"))
        await asyncio.sleep(15)
        return
    except steamgifts.UserSuspended:
        module_data = utils.ModuleData(error=_("User is suspended."))

        async for data in utils.timed_module_data(18000, module_data):
            yield data

        return
    except steamgifts.PrivateProfile:
        yield utils.ModuleData(error=_("Your profile must be public to use steamgifts."), info=_("Waiting Changes"))
        await asyncio.sleep(30)
        return
    except login.LoginError:
        yield utils.ModuleData(error=_("User is not logged in. Trying again in 30 seconds"))
        await asyncio.sleep(30)
        return

    try:
        await steamgifts_session.configure()
    except aiohttp.ClientError:
        yield utils.ModuleData(error=_("Check your connection. (server down?)"))
        await asyncio.sleep(15)
        return
    except steamgifts.ConfigureError:
        yield utils.ModuleData(error=_("Unable to configure steamgifts."))
        await asyncio.sleep(20)
        return

    pinned = config.parser.getboolean("steamgifts", "developer_giveaways")
    points_to_preserve = config.parser.getint("steamgifts", "minimum_points")
    mode = config.parser.get("steamgifts", "mode")
    wait_after_each_strategy = config.parser.getint("steamgifts", "wait_after_each_strategy")
    wait_after_full_cycle = config.parser.getint("steamgifts", "wait_after_full_cycle")

    for strategy_index in range(1, 6):
        strategy = f"steamgifts_strategy{strategy_index}"
        enabled = config.parser.get(strategy, "enable")

        if not enabled:
            yield utils.ModuleData(info=_("Strategy {} is disabled. Skipping.").format(strategy_index))
            continue

        type_ = config.parser.get(strategy, "restrict_type")
        minimum_points = config.parser.getint(strategy, "minimum_points")
        maximum_points = config.parser.getint(strategy, "maximum_points")
        minimum_level = config.parser.getint(strategy, "minimum_level")
        maximum_level = config.parser.getint(strategy, "maximum_level")
        minimum_copies = config.parser.getint(strategy, "minimum_copies")
        maximum_copies = config.parser.getint(strategy, "maximum_copies")
        minimum_metascore = config.parser.getint(strategy, "minimum_metascore")
        maximum_metascore = config.parser.getint(strategy, "maximum_metascore")
        minimum_entries = config.parser.getint(strategy, "minimum_entries")
        maximum_entries = config.parser.getint(strategy, "maximum_entries")

        max_ban_wait = random.randint(5, 15)
        async for data in utils.timed_module_data(max_ban_wait, utils.ModuleData()):
            yield data

        try:
            giveaways = await steamgifts_session.get_giveaways(
                type_,
                (minimum_metascore, maximum_metascore),
                (minimum_level, maximum_level),
                (minimum_entries, maximum_entries),
                (minimum_points, maximum_points),
                (minimum_copies, maximum_copies),
                pinned_giveaways=pinned,
            )
        except aiohttp.ClientError:
            yield utils.ModuleData(error=_("Check your connection. (server down?)"))
            await asyncio.sleep(15)
            return

        wait_enabled = False

        if giveaways:
            sort_type = config.parser.get(strategy, "sort_type")
            sort_name = sort_type[:-1]
            sort_direction = sort_type[-1]

            giveaways = sorted(
                giveaways,
                key=lambda giveaway_: getattr(giveaway_, sort_name),
                reverse=True if sort_direction == '-' else False,
            )
        else:
            yield utils.ModuleData(status=_("No giveaways to join for strategy {}. Skipping.").format(strategy_index))
            continue

        restart = False

        for index, giveaway in enumerate(giveaways):
            module_data = utils.ModuleData(display=giveaway.id, info=giveaway.name)
            max_ban_wait = random.randint(5, 15)

            async for data in utils.timed_module_data(max_ban_wait, module_data):
                yield data

            if steamgifts_session.user_info.points <= points_to_preserve:
                yield utils.ModuleData(status=_("Minimum points reached."))
                wait_enabled = True

                if mode == 'stop_after_minimum_and_restart':
                    restart = True

                break

            yield utils.ModuleData(level=(index, len(giveaway)))

            try:
                if await steamgifts_session.join(giveaway):
                    yield utils.ModuleData(
                        display=giveaway.id,
                        status=f"{_('Joined')} {giveaway.name} "
                               f"(C:{giveaway.copies} P:{giveaway.points} L:{giveaway.level})",
                    )
                    wait_enabled = True
                else:
                    yield utils.ModuleData(display=giveaway.id, error=_("Unable to join {}."))
                    await asyncio.sleep(5)
                    continue
            except aiohttp.ClientError:
                yield utils.ModuleData(error=_("Check your connection. (server down?)"))
                await asyncio.sleep(15)
                wait_enabled = False
                break
            except steamgifts.NoGiveawaysError:
                yield utils.ModuleData(error=_("No giveaways available to join."))
                await asyncio.sleep(15)
                continue
            except steamgifts.GiveawayEndedError:
                yield utils.ModuleData(error=_("Giveaway is already ended."))
                await asyncio.sleep(5)
                continue
            except login.LoginError:
                yield utils.ModuleData(error=_("Login is lost. Trying to relogin."))
                await asyncio.sleep(5)
                wait_enabled = False
                break
            except steamgifts.NoLevelError:
                yield utils.ModuleData(error=_("User don't have required level to join."))
                await asyncio.sleep(5)
                continue
            except steamgifts.NoPointsError:
                yield utils.ModuleData(error=_("User don't have required points to join."))
                await asyncio.sleep(5)

                if steamgifts_session.user_info.points <= 2:
                    break

                continue

        if not wait_enabled:
            await asyncio.sleep(10)
            return

        wait_offset = random.randint(
            wait_after_each_strategy,
            wait_after_each_strategy + int(wait_after_each_strategy / 6),
        )

        module_data = utils.ModuleData(info=_("Waiting before next strategy"))

        async for data in utils.timed_module_data(wait_offset, module_data):
            yield data

        if restart:
            yield utils.ModuleData(info=_("Restarting due to mode selection"))
            break

    wait_offset = random.randint(
        wait_after_full_cycle,
        wait_after_full_cycle + int(wait_after_full_cycle / 6),
    )

    module_data = utils.ModuleData(info=_("Waiting for next cycle"))

    async for data in utils.timed_module_data(wait_offset, module_data):
        yield data



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#

import asyncio
import contextlib
import logging
from gi.repository import Gio, Gtk, Gdk
from subprocess import call
from typing import Union, Optional, Tuple, Any, List

import stlib
from stlib import login, universe, plugins
from . import confirmation, utils, coupon
from .authenticator import NewAuthenticatorDialog
from .login import LoginDialog
from .. import config, i18n, core

_ = i18n.get_translation
log = logging.getLogger(__name__)

if stlib.steamworks_available:
    from stlib import client


# noinspection PyUnusedLocal
class Main(Gtk.ApplicationWindow):
    def __init__(self, application: Gtk.Application, title: str) -> None:
        super().__init__(application=application, title=title)
        self.application = application

        _display = Gdk.Display.get_default()
        _style_provider = Gtk.CssProvider()

        _style_provider.load_from_data(
            b"* { border-radius: 2px; }"
            b"label.warning { background-color: darkblue; color: white; }"
            b"label.critical { background-color: darkred; color: white; }"
        )

        _style_context = self.get_style_context()

        _style_context.add_provider_for_display(
            _display,
            _style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        header_bar = Gtk.HeaderBar()

        menu = Gio.Menu()
        menu.append(_("Settings"), "app.settings")
        menu.append(_("About"), "app.about")
        menu.append(_("Exit"), "app.exit")

        menu_button = Gtk.MenuButton()
        menu_button.set_icon_name("open-menu")
        menu_button.set_menu_model(menu)
        header_bar.pack_end(menu_button)

        self.set_default_size(650, 10)
        self.set_resizable(False)

        if config.parser.getboolean("general", "show_close_button"):
            self.set_deletable(True)
        else:
            self.set_deletable(False)

        self.set_titlebar(header_bar)
        self.set_title('Steam Tools NG')

        main_grid = Gtk.Grid()
        main_grid.set_margin_start(10)
        main_grid.set_margin_end(10)
        main_grid.set_margin_top(10)
        main_grid.set_margin_bottom(10)
        self.set_child(main_grid)

        self.user_info_label = Gtk.Label()
        self.user_info_label.set_halign(Gtk.Align.END)
        header_bar.pack_start(self.user_info_label)

        main_tabs = Gtk.Stack()
        main_tabs.set_hhomogeneous(True)
        main_grid.attach(main_tabs, 1, 2, 1, 1)

        switcher = Gtk.StackSwitcher()
        switcher.set_stack(main_tabs)
        main_grid.attach(switcher, 1, 1, 1, 1)

        steamguard_section = utils.Section("steamguard")
        steamguard_section.stackup_section("SteamGuard", main_tabs)

        cardfarming_section = utils.Section("cardfarming")
        cardfarming_section.stackup_section("CardFarming", main_tabs)

        steamgifts_section = utils.Section("steamgifts")
        steamgifts_section.stackup_section("SteamGifts", main_tabs)

        steamtrades_section = utils.Section("steamtrades")
        steamtrades_section.stackup_section("SteamTrades", main_tabs)

        coupons_section = utils.Section("coupons")
        coupons_section.stackup_section(_("Coupons"), main_tabs)

        # grid managed by plugin switch
        self.steamguard_status = utils.Status(4)
        self.cardfarming_status = utils.Status(6)
        self.steamgifts_status = utils.Status(5)
        self.steamtrades_status = utils.Status(5)

        steamguard_section.grid.attach(self.steamguard_status, 0, 0, 2, 1)

        steamguard_stack = Gtk.Stack()
        steamguard_stack.set_vexpand(True)
        steamguard_section.grid.attach(steamguard_stack, 1, 1, 1, 1)

        steamguard_sidebar = Gtk.StackSidebar()
        steamguard_sidebar.set_stack(steamguard_stack)
        steamguard_sidebar.set_size_request(150, -1)
        steamguard_section.grid.attach(steamguard_sidebar, 0, 1, 1, 1)

        self.confirmations_grid = Gtk.Grid()
        self.confirmations_grid.set_row_spacing(10)
        steamguard_stack.add_titled(self.confirmations_grid, "confirmations", _("Confirmations"))

        self.confirmations_tree = utils.SimpleTextTree(
            _('confid'), _('creatorid'), _('key'), _('give'), _('to'), _('receive'),
            overlay_scrolling=False,
        )

        self.confirmations_grid.attach(self.confirmations_tree, 0, 0, 4, 1)

        for index, column in enumerate(self.confirmations_tree.view.get_columns()):
            if index in (0, 1, 2):
                column.set_visible(False)

            if index == 4:
                column.set_fixed_width(100)
            else:
                column.set_fixed_width(200)

        self.confirmations_tree.view.set_has_tooltip(True)
        self.confirmations_tree.view.connect('query-tooltip', self.on_query_confirmations_tooltip)

        confirmation_tree_selection = self.confirmations_tree.view.get_selection()
        confirmation_tree_selection.connect("changed", self.on_tree_selection_changed)

        accept_button = Gtk.Button()
        accept_button.set_margin_start(3)
        accept_button.set_margin_end(3)
        accept_button.set_label(_('Accept'))
        accept_button.connect('clicked', self.on_validate_confirmations, "allow", confirmation_tree_selection)
        self.confirmations_grid.attach(accept_button, 0, 1, 1, 1)

        cancel_button = Gtk.Button()
        cancel_button.set_margin_start(3)
        cancel_button.set_margin_end(3)
        cancel_button.set_label(_('Cancel'))
        cancel_button.connect('clicked', self.on_validate_confirmations, "cancel", confirmation_tree_selection)
        self.confirmations_grid.attach(cancel_button, 1, 1, 1, 1)

        accept_all_button = Gtk.Button()
        accept_all_button.set_margin_start(3)
        accept_all_button.set_margin_end(3)
        accept_all_button.set_label(_('Accept All'))
        accept_all_button.connect('clicked', self.on_validate_confirmations, "allow", self.confirmations_tree.store)
        self.confirmations_grid.attach(accept_all_button, 2, 1, 1, 1)

        cancel_all_button = Gtk.Button()
        cancel_all_button.set_margin_start(3)
        cancel_all_button.set_margin_end(3)
        cancel_all_button.set_label(_('Cancel All'))
        cancel_all_button.connect('clicked', self.on_validate_confirmations, "cancel", self.confirmations_tree.store)
        self.confirmations_grid.attach(cancel_all_button, 3, 1, 1, 1)

        steamguard_settings = utils.Section("steamguard")
        steamguard_settings.stackup_section(_("Settings"), steamguard_stack)
        steamguard_settings.grid.set_halign(Gtk.Align.CENTER)

        steamguard_enable = steamguard_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 0)
        steamguard_enable.set_margin_top(40)
        steamguard_enable.label.set_margin_top(40)
        steamguard_enable.set_halign(Gtk.Align.END)
        steamguard_enable.connect("state-set", utils.on_setting_state_set)

        if not config.parser.get("login", "shared_secret"):
            self.steamguard_status.set_sensitive(False)
            steamguard_enable.set_active(False)
            _steamguard_disabled = Gtk.Label()
            _steamguard_disabled.set_justify(Gtk.Justification.CENTER)
            _steamguard_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "steamguard module has been disabled because you have\n"
                "logged in but no shared secret is found. To enable it again,\n"
                "go to Advanced and add a valid shared secret\n"
                "or use STNG as your Steam Authenticator\n"
            )

            _steamguard_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            steamguard_section.grid.attach(_steamguard_disabled, 0, 0, 2, 1)

        confirmations_enable = steamguard_settings.new_item(
            "enable_confirmations", _("Enable Confirmations:"),
            Gtk.Switch,
            0, 1,
        )

        confirmations_enable.set_halign(Gtk.Align.END)
        confirmations_enable.connect("state-set", utils.on_setting_state_set)

        if not config.parser.get("login", "identity_secret"):
            self.confirmations_grid.set_sensitive(False)
            confirmations_enable.set_active(False)
            _confirmations_disabled = Gtk.Label()
            _confirmations_disabled.set_justify(Gtk.Justification.CENTER)
            _confirmations_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "confirmations module has been disabled because you have\n"
                "logged in but no identity secret is found. To enable it again,\n"
                "go to login -> advanced and add a valid identity secret\n"
                "or use STNG as your Steam Authenticator\n"
            )

            _confirmations_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            self.confirmations_grid.attach(_confirmations_disabled, 0, 0, 4, 1)

        login_button = Gtk.Button()
        login_button.set_margin_top(40)
        login_button.set_label(_("Login with another account"))
        login_button.set_name("login_button")
        login_button.connect('clicked', self.on_login_button_clicked)
        steamguard_settings.grid.attach(login_button, 0, 2, 2, 1)

        new_authenticator_button = Gtk.Button()
        new_authenticator_button.set_label(_("Use STNG as your Steam Authenticator"))
        new_authenticator_button.set_name("new_authenticator_button")
        new_authenticator_button.connect("clicked", self.on_new_authenticator_clicked)
        steamguard_settings.grid.attach(new_authenticator_button, 0, 3, 2, 1)

        reset_password_button = Gtk.Button()
        reset_password_button.set_label(_("Remove Saved Password"))
        reset_password_button.set_name("reset_password_button")
        reset_password_button.connect("clicked", self.on_reset_password_clicked)
        steamguard_settings.grid.attach(reset_password_button, 0, 4, 2, 1)

        steamguard_advanced = utils.Section("login")
        steamguard_advanced.stackup_section(_("Advanced"), steamguard_stack, scroll=True)

        warning_label = Gtk.Label()
        warning_label.set_markup(utils.markup(
            _("Warning: Don't mess up these settings unless you know what you are doing!"),
            color='darkred' if self.theme == 'light' else 'red',
        ))
        steamguard_advanced.grid.attach(warning_label, 0, 0, 2, 1)

        shared_secret = steamguard_advanced.new_item('shared_secret', _("Shared Secret:"), Gtk.Entry, 0, 1)
        shared_secret.connect('changed', utils.on_setting_changed)

        token_item = steamguard_advanced.new_item("token", _("Token:"), Gtk.Entry, 0, 2)
        token_item.connect("changed", utils.on_setting_changed)

        token_secure_item = steamguard_advanced.new_item("token_secure", _("Token Secure:"), Gtk.Entry, 0, 3)
        token_secure_item.connect("changed", utils.on_setting_changed)

        identity_secret = steamguard_advanced.new_item('identity_secret', _("Identity Secret:"), Gtk.Entry, 0, 4)
        identity_secret.connect('changed', utils.on_setting_changed)

        deviceid = steamguard_advanced.new_item('deviceid', _("Device ID:"), Gtk.Entry, 0, 5)
        deviceid.connect('changed', utils.on_setting_changed)

        steamid_item = steamguard_advanced.new_item("steamid", _("Steam ID:"), Gtk.Entry, 0, 6)
        steamid_item.set_input_purpose(Gtk.InputPurpose.DIGITS)
        steamid_item.connect("changed", utils.on_digit_only_setting_changed)

        account_name = steamguard_advanced.new_item('account_name', _("Username:"), Gtk.Entry, 0, 7)
        account_name.connect('changed', utils.on_setting_changed)

        reset_button = utils.AsyncButton()
        reset_button.set_label(_("Reset Everything (USE WITH CAUTION!!!)"))
        reset_button.set_name("reset_button")
        reset_button.connect("clicked", self.on_reset_clicked)
        steamguard_advanced.grid.attach(reset_button, 0, 8, 2, 1)

        cardfarming_section.grid.attach(self.cardfarming_status, 0, 0, 2, 1)

        cardfarming_stack = Gtk.Stack()
        cardfarming_stack.set_vexpand(True)
        cardfarming_section.grid.attach(cardfarming_stack, 1, 1, 1, 1)

        cardfarming_sidebar = Gtk.StackSidebar()
        cardfarming_sidebar.set_stack(cardfarming_stack)
        cardfarming_sidebar.set_size_request(150, -1)
        cardfarming_section.grid.attach(cardfarming_sidebar, 0, 1, 1, 1)

        cardfarming_settings = utils.Section("cardfarming")
        cardfarming_settings.stackup_section(_("Settings"), cardfarming_stack, scroll=True)

        cardfarming_enable = cardfarming_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 1)
        cardfarming_enable.set_halign(Gtk.Align.END)
        cardfarming_enable.connect("state-set", utils.on_setting_state_set)

        mandatory_waiting = cardfarming_settings.new_item("mandatory_waiting", _("Mandatory waiting:"), Gtk.Entry, 0, 2)
        mandatory_waiting.connect("changed", utils.on_digit_only_setting_changed)

        wait_while_running = cardfarming_settings.new_item(
            "wait_while_running",
            _("Wait while running:"),
            Gtk.Entry,
            0, 3,
        )
        wait_while_running.connect("changed", utils.on_digit_only_setting_changed)

        wait_for_drops = cardfarming_settings.new_item("wait_for_drops", _("Wait for drops:"), Gtk.Entry, 0, 4)
        wait_for_drops.connect("changed", utils.on_digit_only_setting_changed)

        max_concurrency = cardfarming_settings.new_item("max_concurrency", _("Max concurrency:"), Gtk.Entry, 0, 5)
        max_concurrency.connect("changed", utils.on_digit_only_setting_changed)

        cardfarming_invisible = cardfarming_settings.new_item("invisible", _("Invisible:"), Gtk.Switch, 0, 6)
        cardfarming_invisible.set_halign(Gtk.Align.END)
        cardfarming_invisible.connect("state-set", utils.on_setting_state_set)

        reverse_sorting = cardfarming_settings.new_item("reverse_sorting", _("Reverse Sorting:"), Gtk.Switch, 0, 7)
        reverse_sorting.set_halign(Gtk.Align.END)
        reverse_sorting.connect("state-set", utils.on_setting_state_set)

        if not stlib.steamworks_available:
            cardfarming_settings.set_sensitive(False)
            cardfarming_enable.set_active(False)
            _cardfarming_disabled = Gtk.Label()
            _cardfarming_disabled.set_justify(Gtk.Justification.CENTER)
            _cardfarming_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "cardfarming module has been disabled because you have\n"
                "a stlib built without SteamWorks support. To enable it again,\n"
                "reinstall stlib with SteamWorks support\n"
            )

            _cardfarming_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            cardfarming_section.attach(_cardfarming_disabled, 0, 0, 2, 1)

        # TODO: Maintain plugin switch?
        steamgifts_section.grid.attach(self.steamgifts_status, 0, 0, 2, 1)

        steamgifts_stack = Gtk.Stack()
        steamgifts_stack.set_vexpand(True)
        steamgifts_section.grid.attach(steamgifts_stack, 1, 1, 1, 1)

        steamgifts_sidebar = Gtk.StackSidebar()
        steamgifts_sidebar.set_stack(steamgifts_stack)
        steamgifts_sidebar.set_size_request(150, -1)
        steamgifts_section.grid.attach(steamgifts_sidebar, 0, 1, 1, 1)

        steamgifts_settings = utils.Section("steamgifts")
        steamgifts_settings.stackup_section(_("Settings"), steamgifts_stack)

        steamgifts_enable = steamgifts_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 0)
        steamgifts_enable.set_halign(Gtk.Align.END)
        steamgifts_enable.connect('state-set', utils.on_setting_state_set)

        developer_giveaways = steamgifts_settings.new_item(
            "developer_giveaways", _("Developer Giveaways"),
            Gtk.Switch,
            0, 1,
        )
        developer_giveaways.set_halign(Gtk.Align.END)
        developer_giveaways.connect("state-set", utils.on_setting_state_set)

        steamgifts_mode = steamgifts_settings.new_item(
            "mode", _("Mode:"),
            Gtk.ComboBoxText,
            0, 2,
            items=config.steamgifts_modes,
        )
        steamgifts_mode.connect("changed", utils.on_combo_setting_changed, config.steamgifts_modes)

        wait_after_each_strategy = steamgifts_settings.new_item(
            "wait_after_each_strategy", _("Wait after each strategy:"),
            Gtk.Entry,
            0, 3,
        )
        wait_after_each_strategy.connect("changed", utils.on_digit_only_setting_changed)

        wait_after_full_cycle = steamgifts_settings.new_item(
            "wait_after_full_cycle", _("Wait after full cycle:"),
            Gtk.Entry,
            0, 4,
        )
        wait_after_full_cycle.connect("changed", utils.on_digit_only_setting_changed)

        minimum_points = steamgifts_settings.new_item("minimum_points", _("Minimum points:"), Gtk.Entry, 0, 5)
        minimum_points.connect("changed", utils.on_digit_only_setting_changed)

        for index in range(1, 6):
            strategy_section = utils.Section(f"steamgifts_strategy{index}")
            strategy_section.stackup_section(_("Strategy {}").format(index), steamgifts_stack, scroll=True)

            label = Gtk.Label()
            label.set_text(_("Strategy {}").format(index))
            strategy_section.grid.attach(label, 0, 0, 1, 1)

            enable = strategy_section.new_item("enable", None, Gtk.Switch, 2, 0)
            enable.set_halign(Gtk.Align.END)
            enable.connect("state-set", utils.on_setting_state_set)

            minimum_label = Gtk.Label()
            minimum_label.set_text(_("Minimum"))
            strategy_section.grid.attach(minimum_label, 1, 1, 1, 1)

            maximum_label = Gtk.Label()
            maximum_label.set_text(_("Maximum"))
            strategy_section.grid.attach(maximum_label, 2, 1, 1, 1)

            for tree_level, item in enumerate(["points", "level", "copies", "metascore", "entries"]):
                label = Gtk.Label()
                label.set_text(_(item))
                strategy_section.grid.attach(label, 0, tree_level + 2, 1, 1)

                minimum = strategy_section.new_item(f"minimum_{item}", None, Gtk.Entry, 1, tree_level + 2)
                minimum.connect("changed", utils.on_digit_only_setting_changed)

                maximum = strategy_section.new_item(f"maximum_{item}", None, Gtk.Entry, 2, tree_level + 2)
                maximum.connect("changed", utils.on_digit_only_setting_changed)

            restrict_type = strategy_section.new_item(
                "restrict_type",
                _("Restrict Type:"),
                Gtk.ComboBoxText,
                0, 7,
                items=config.giveaway_types,
            )
            restrict_type.connect("changed", utils.on_combo_setting_changed, config.giveaway_types)

            sort_type = strategy_section.new_item(
                "sort_type",
                _("Sort Type:"),
                Gtk.ComboBoxText,
                0, 8,
                items=config.giveaway_sort_types,
            )
            sort_type.connect("changed", utils.on_combo_setting_changed, config.giveaway_sort_types)

            # setattr(self, f"steamgifts_strategy{index}", strategy_section)

        if not plugins.has_plugin("steamgifts"):
            steamgifts_section.set_sensitive(False)
            steamgifts_enable.set_active(False)
            _steamgifts_disabled = Gtk.Label()
            _steamgifts_disabled.set_justify(Gtk.Justification.CENTER)
            _steamgifts_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "steamgifts module has been disabled because you don't\n"
                "have steamgifts plugin installed. To enable it again,\n"
                "install the steamgifts plugin.\n"
            )

            _steamgifts_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            steamgifts_section.grid.attach(_steamgifts_disabled, 0, 0, 2, 2)

        steamtrades_section.grid.attach(self.steamtrades_status, 0, 0, 2, 1)

        steamtrades_stack = Gtk.Stack()
        steamtrades_stack.set_vexpand(True)
        steamtrades_section.grid.attach(steamtrades_stack, 1, 1, 1, 1)

        steamtrades_sidebar = Gtk.StackSidebar()
        steamtrades_sidebar.set_stack(steamtrades_stack)
        steamtrades_sidebar.set_size_request(150, -1)
        steamtrades_section.grid.attach(steamtrades_sidebar, 0, 1, 1, 1)

        steamtrades_settings = utils.Section("steamtrades")
        steamtrades_settings.stackup_section(_("Settings"), steamtrades_stack)

        steamtrades_enable = steamtrades_settings.new_item("enable", _("Enable:"), Gtk.Switch, 0, 1)
        steamtrades_enable.label.set_margin_top(40)
        steamtrades_enable.set_margin_top(40)
        steamtrades_enable.set_halign(Gtk.Align.END)
        steamtrades_enable.connect("state-set", utils.on_setting_state_set)

        trade_ids = steamtrades_settings.new_item("trade_ids", _("Trade IDs:"), Gtk.Entry, 0, 2)
        trade_ids.set_placeholder_text('12345, asdfg, ...')
        trade_ids.connect("changed", utils.on_setting_changed)

        wait_for_bump = steamtrades_settings.new_item("wait_for_bump", _("Wait for Bump:"), Gtk.Entry, 0, 3)
        wait_for_bump.connect("changed", utils.on_digit_only_setting_changed)

        if not plugins.has_plugin("steamtrades"):
            steamtrades_settings.set_sensitive(False)
            steamtrades_enable.set_active(False)
            _steamtrades_disabled = Gtk.Label()
            _steamtrades_disabled.set_justify(Gtk.Justification.CENTER)
            _steamtrades_disabled.set_halign(Gtk.Align.CENTER)

            _message = _(
                "steamtrades module has been disabled because you don't\n"
                "have steamtrades plugin installed. To enable it again,\n"
                "install the steamtrades plugin.\n"
            )

            _steamtrades_disabled.set_markup(utils.markup(_message, color="hotpink", background="black"))
            steamtrades_settings.grid.attach(_steamtrades_disabled, 0, 1, 2, 2)

        self.coupon_warning = Gtk.Label()
        self.coupon_warning.set_markup(utils.markup(
            _("Warning: It's a heavy uncached operation. Fetch only once a day or you will be blocked."),
            color='darkred' if self.theme == 'light' else 'red',
        ))
        self.coupon_warning.set_margin_top(37)
        self.coupon_warning.set_margin_bottom(37)
        coupons_section.grid.attach(self.coupon_warning, 0, 0, 2, 1)

        coupons_stack = Gtk.Stack()
        coupons_stack.set_vexpand(True)
        coupons_section.grid.attach(coupons_stack, 1, 1, 1, 1)

        coupons_sidebar = Gtk.StackSidebar()
        coupons_sidebar.set_stack(coupons_stack)
        coupons_sidebar.set_size_request(150, -1)
        coupons_section.grid.attach(coupons_sidebar, 0, 1, 1, 1)

        self.coupons_grid = Gtk.Grid()
        self.coupons_grid.set_row_spacing(10)
        coupons_stack.add_titled(self.coupons_grid, "coupons_list", _("Coupon List"))

        self.coupons_tree = utils.SimpleTextTree(
            _('price'), _('name'), 'link', 'botid', 'token', 'assetid',
            overlay_scrolling=False,
            model=Gtk.ListStore,
        )

        self.coupons_tree.store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self.coupons_tree.store.set_sort_func(0, self.coupon_sorting)
        self.coupons_grid.attach(self.coupons_tree, 0, 0, 4, 2)

        for index, column in enumerate(self.coupons_tree.view.get_columns()):
            if index in (2, 3, 4, 5):
                column.set_visible(False)

        self.coupons_tree.view.connect('row-activated', self.on_coupon_double_clicked)

        coupon_tree_selection = self.coupons_tree.view.get_selection()
        coupon_tree_selection.connect("changed", self.on_tree_selection_changed)

        self.coupon_progress = Gtk.LevelBar()
        self.coupons_grid.attach(self.coupon_progress, 0, 3, 4, 1)

        fetch_coupons_button = Gtk.Button()
        fetch_coupons_button.set_margin_start(3)
        fetch_coupons_button.set_margin_end(3)
        fetch_coupons_button.set_label(_('Fetch'))
        fetch_coupons_button.connect('clicked', self.on_fetch_coupons)
        self.coupons_grid.attach(fetch_coupons_button, 0, 4, 1, 1)

        self.fetch_coupon_event = asyncio.Event()

        stop_fetching_coupons_button = Gtk.Button()
        stop_fetching_coupons_button.set_margin_start(3)
        stop_fetching_coupons_button.set_margin_end(3)
        stop_fetching_coupons_button.set_label(_('Stop fetching'))
        stop_fetching_coupons_button.connect('clicked', self.on_stop_fetching_coupons)
        self.coupons_grid.attach(stop_fetching_coupons_button, 1, 4, 1, 1)

        get_coupon_button = Gtk.Button()
        get_coupon_button.set_margin_start(3)
        get_coupon_button.set_margin_end(3)
        get_coupon_button.set_label(_('Get selected'))
        get_coupon_button.connect('clicked', self.on_coupon_action, coupon_tree_selection)
        self.coupons_grid.attach(get_coupon_button, 2, 4, 1, 1)

        give_coupon_button = Gtk.Button()
        give_coupon_button.set_margin_start(3)
        give_coupon_button.set_margin_end(3)
        give_coupon_button.set_label(_('Send'))
        give_coupon_button.connect('clicked', self.on_coupon_action)
        self.coupons_grid.attach(give_coupon_button, 3, 4, 1, 1)

        coupons_settings = utils.Section("coupons")
        coupons_settings.stackup_section(_("Settings"), coupons_stack)

        coupon_botids = coupons_settings.new_item("botids", _("BotIDs:"), Gtk.Entry, 0, 1)
        coupon_botids.set_placeholder_text('12345, asdfg, ...')
        coupon_botids.connect("changed", utils.on_setting_changed)

        coupon_tokens = coupons_settings.new_item("tokens", _("Tokens:"), Gtk.Entry, 0, 2)
        coupon_tokens.set_placeholder_text('12345, asdfg, ...')
        coupon_tokens.connect("changed", utils.on_setting_changed)

        coupon_botid_to_donate = coupons_settings.new_item("botid_to_donate", _("BotID To Donate:"), Gtk.Entry, 0, 3)
        coupon_botid_to_donate.connect("changed", utils.on_digit_only_setting_changed)

        coupon_token_to_donate = coupons_settings.new_item("token_to_donate", _("Token To Donate:"), Gtk.Entry, 0, 4)
        coupon_token_to_donate.connect("changed", utils.on_setting_changed)

        coupon_blacklist = coupons_settings.new_item("blacklist", _("Blacklist:"), Gtk.Entry, 0, 5)
        coupon_blacklist.connect("changed", utils.on_setting_changed)

        coupon_discount = coupons_settings.new_item(
            "minimum_discount",
            _("Minimum Discount:"),
            Gtk.ComboBoxText,
            0, 6,
            items=config.coupon_discounts,
        )
        coupon_discount.connect("changed", utils.on_combo_setting_changed, config.coupon_discounts)

        self.statusbar = utils.StatusBar()
        main_grid.attach(self.statusbar, 1, 3, 1, 1)

        self.connect("destroy", self.application.on_exit_activate)
        self.connect("close-request", self.application.on_exit_activate)

        loop = asyncio.get_event_loop()

        plugin_status_task = loop.create_task(self.plugin_status())
        plugin_status_task.add_done_callback(utils.safe_task_callback)

        user_info_task = loop.create_task(self.user_info())
        user_info_task.add_done_callback(utils.safe_task_callback)

    @property
    def theme(self) -> str:
        return config.parser.get('general', 'theme')

    async def user_info(self) -> None:
        while self.get_realized():
            account_name = config.parser.get('login', 'account_name')
            steamid_raw = config.parser.get('login', 'steamid')

            try:
                steamid = universe.generate_steamid(steamid_raw)
            except ValueError:
                log.warning(_("SteamId is invalid"))
                steamid = None

            login_session = None

            with contextlib.suppress(IndexError):
                login_session = login.Login.get_session(0)

            if not steamid or not login_session or not await login_session.is_logged_in(steamid):
                self.application.main_window.user_info_label.set_markup(
                    utils.markup(
                        _('Not logged in'),
                        color='darkred' if self.theme == 'light' else 'red',
                        size='small',
                    )
                )

                await asyncio.sleep(5)
                continue

            self.application.main_window.user_info_label.set_markup(
                utils.markup(
                    _('You are logged in as:\n'),
                    color='darkgreen' if self.theme == 'light' else 'green',
                    size='small',
                ) +
                utils.markup(
                    account_name,
                    color='darkblue' if self.theme == 'light' else 'blue',
                    size='small',
                ) +
                utils.markup(
                    f" {self.application.steamid.id3_string}",
                    color='grey',
                    size='small',
                )
            )

            await asyncio.sleep(30)

    async def plugin_status(self) -> None:
        while self.get_realized():
            for plugin_name in config.plugins.keys():
                if plugin_name in ["coupons", "confirmations"]:
                    if plugin_name == "confirmations":
                        enabled = config.parser.getboolean("steamguard", "enable_confirmations")
                    else:
                        enabled = config.parser.getboolean(plugin_name, "enable")

                    main = getattr(self, f'{plugin_name}_grid')
                    tree = getattr(self, f'{plugin_name}_tree')

                    if enabled:
                        tree.disabled = False
                        main.set_sensitive(True)
                    else:
                        tree.disabled = True
                        main.set_sensitive(False)
                else:
                    enabled = config.parser.getboolean(plugin_name, "enable")
                    status = getattr(self, f'{plugin_name}_status')

                    if not enabled:
                        await asyncio.sleep(5)
                        status.set_status(_("Disabled"))
                        status.set_info("")

            await asyncio.sleep(3)

    @staticmethod
    def on_query_confirmations_tooltip(
            tree_view: Gtk.TreeView,
            x_coord: int,
            y_coord: int,
            keyboard_tip: bool,
            tooltip: Gtk.Tooltip,
    ) -> bool:
        context = tree_view.get_tooltip_context(x_coord, y_coord, keyboard_tip)

        if context[0]:
            if context.model.iter_depth(context.iter) != 0:
                return False

            tooltip.set_text(
                f'ConfID:{context.model.get_value(context.iter, 0)}\n'
                f'CreatorID{context.model.get_value(context.iter, 1)}\n'
                f'Key:{context.model.get_value(context.iter, 2)}'
            )

            return True

        return False

    def on_fetch_coupons(self, button: Gtk.Button) -> None:
        self.fetch_coupon_event.set()

    def on_stop_fetching_coupons(self, button: Gtk.Button) -> None:
        self.fetch_coupon_event.clear()

    def on_coupon_action(self, button: Gtk.Button, model: Union[Gtk.TreeModel, Gtk.TreeSelection] = None) -> None:
        if model:
            coupon_dialog = coupon.CouponDialog(self, self.application, *model.get_selected())
        else:
            coupon_dialog = coupon.CouponDialog(self, self.application)

        coupon_dialog.show()

    def on_validate_confirmations(
            self,
            button: Gtk.Button,
            action: str,
            model: Union[Gtk.TreeModel, Gtk.TreeSelection]) -> None:
        if isinstance(model, Gtk.TreeModel):
            finalize_dialog = confirmation.FinalizeDialog(
                self,
                self.application,
                action,
                model,
                False
            )
        else:
            finalize_dialog = confirmation.FinalizeDialog(
                self,
                self.application,
                action,
                *model.get_selected()
            )

        finalize_dialog.show()

    @staticmethod
    def on_coupon_double_clicked(view: Gtk.TreeView, path: Gtk.TreePath, column: Gtk.TreeViewColumn) -> None:
        model = view.get_model()
        url = model[path][2]
        steam_running = False

        if stlib.steamworks_available:
            with contextlib.suppress(ProcessLookupError):
                with client.SteamGameServer() as server:
                    steam_running = True

        if steam_running:
            url = f"steam://openurl/{url}"

        call(f'{config.file_manager} "{url}"')

    @staticmethod
    def on_tree_selection_changed(selection: Gtk.TreeSelection) -> None:
        model, iter_ = selection.get_selected()

        if iter_:
            parent = model.iter_parent(iter_)

            if parent:
                selection.select_iter(parent)

    @staticmethod
    def coupon_sorting(model: Gtk.TreeModel, iter1: Gtk.TreeIter, iter2: Gtk.TreeIter, user_data: Any) -> Any:
        column, _ = model.get_sort_column_id()
        price1 = model.get_value(iter1, column)
        price2 = model.get_value(iter2, column)

        if float(price1) < float(price2):
            return -1

        if price1 == price2:
            return 0

        return 1

    def set_status(
            self,
            module: str,
            module_data: Optional[core.utils.ModuleData] = None,
            *,
            display: str = '',
            status: str = '',
            info: str = '',
            error: str = '',
            level: Tuple[int, int] = (0, 0),
            suppress_logging: bool = False,
    ) -> None:
        _status = getattr(self, f'{module}_status')

        if not module_data:
            module_data = core.utils.ModuleData(display, status, info, error, level, suppress_logging=suppress_logging)

        if module_data.display:
            if not module_data.suppress_logging:
                log.debug(f"display data: {module_data.display}")

            _status.set_display(module_data.display)
        else:
            _status.unset_display()

        if module_data.status:
            if not module_data.suppress_logging:
                log.debug(f"status data: {module_data.status}")

            _status.set_status(module_data.status)

        if module_data.info:
            if not module_data.suppress_logging:
                log.info(module_data.info)

            _status.set_info(module_data.info)

        if module_data.error:
            if not module_data.suppress_logging:
                log.error(module_data.error)

            _status.set_error(module_data.error)

        if module_data.level:
            _status.set_level(*module_data.level)

    def get_play_event(self, module: str) -> asyncio.Event:
        _status = getattr(self, f'{module}_status')
        assert isinstance(_status, utils.Status)
        return _status.play_event

    async def on_reset_clicked(self, button: Gtk.Button) -> None:
        login_dialog = LoginDialog(self, self.application)
        login_dialog.status.info(_("Reseting... Please wait!"))
        login_dialog.set_deletable(False)
        login_dialog.user_details_section.hide()
        login_dialog.advanced_login.hide()
        login_dialog.show()
        await asyncio.sleep(3)

        config.config_file.unlink(missing_ok=True)

        config.parser.clear()
        config.init()
        self.destroy()

    def on_login_button_clicked(self, button: Gtk.Button) -> None:
        login_dialog = LoginDialog(self, self.application)
        login_dialog.shared_secret_item.set_text('')
        login_dialog.identity_secret_item.set_text('')
        login_dialog.show()

    def on_new_authenticator_clicked(self, button: Gtk.Button) -> None:
        new_authenticator_dialog = NewAuthenticatorDialog(self, self.application)
        new_authenticator_dialog.show()

    def on_reset_password_clicked(self, button: Gtk.Button) -> None:
        login_dialog = LoginDialog(self, self.application)
        login_dialog.status.info(_("Removing saved password..."))
        login_dialog.user_details_section.hide()
        login_dialog.advanced_login.hide()
        login_dialog.set_deletable(False)
        login_dialog.show()

        config.new("login", "password", "")

        def reset_password_callback() -> None:
            login_dialog.destroy()
            self.destroy()

        login_dialog.status.info(_("Successful!\nExiting..."))
        asyncio.get_event_loop().call_later(3, reset_password_callback)



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import aiohttp
import logging
from typing import AsyncGenerator

from stlib import login, universe, community
from . import utils
from .. import i18n, config

_ = i18n.get_translation
log = logging.getLogger(__name__)


async def main(steamid: universe.SteamId) -> AsyncGenerator[utils.ModuleData, None]:
    identity_secret = config.parser.get("login", "identity_secret")
    session = community.Community.get_session(0)

    if not identity_secret:
        config.new("confirmations", "enable", "false")
        module_data = utils.ModuleData(error=_("The current identity secret is invalid."), info=_("Waiting Changes"))

        async for data in utils.timed_module_data(10, module_data):
            yield data

        return

    deviceid = config.parser.get("login", "deviceid")

    if not deviceid:
        log.warning(_("Unable to find deviceid. Generating from identity."))
        deviceid = universe.generate_device_id(identity_secret)
        config.new("login", "deviceid", deviceid)

    try:
        confirmations = await session.get_confirmations(identity_secret, steamid, deviceid)
    except AttributeError as exception:
        log.error(str(exception))
        module_data = utils.ModuleData(error=_("Error when fetch confirmations"), info=_("Waiting Changes"))
    except ProcessLookupError:
        module_data = utils.ModuleData(error=_("Steam is not running"), info=_("Waiting Changes"))
    except login.LoginError:
        module_data = utils.ModuleData(error=_("Not logged in"), action="login")
    except aiohttp.ClientError:
        module_data = utils.ModuleData(error=_("Check your connection. (server down?)"), info=_("Waiting Changes"))
    else:
        module_data = utils.ModuleData(action="update", raw_data=confirmations)

    async for data in utils.timed_module_data(20, module_data):
        yield data



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
from collections import OrderedDict
from xml.etree import ElementTree

import asyncio
import functools
import html
import logging
import traceback
from gi.repository import Gtk, Gdk
from traceback import StackSummary
from types import FrameType
from typing import Any, Callable, List, Optional, Union, Type, Tuple

from stlib import internals
from . import async_gtk
from .. import i18n, config

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class ClickableLabel(Gtk.Label):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self.set_hexpand(True)

        pointer = Gdk.Cursor.new_from_name('pointer')
        self.set_cursor(pointer)

        self.gesture = Gtk.GestureClick()
        self.add_controller(self.gesture)

    def connect(self, signal: str, callback: Callable[..., Any]) -> None:
        if signal == 'clicked':
            self.gesture.connect('pressed', callback)
        else:
            raise NotImplementedError


class VariableButton(Gtk.Button):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        super().connect('clicked', self.__callback)
        self._user_callback: Optional[Callable[..., Any]] = None
        self._user_args: Any = None
        self._user_kwargs: Any = None

    def __callback(self, button: Gtk.Button) -> None:
        if not self._user_callback:
            return

        self._user_callback(button, *self._user_args, **self._user_kwargs)

    def connect(self, signal: str, callback: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        if signal != "clicked":
            raise NotImplementedError

        self._user_callback = callback
        self._user_args = args
        self._user_kwargs = kwargs


class AsyncButton(Gtk.Button):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        super().connect('clicked', self.__callback)
        self._user_callback: Optional[Callable[..., Any]] = None
        self._user_args: Any = None
        self._user_kwargs: Any = None

    def __callback(self, button: Gtk.Button) -> None:
        if not self._user_callback:
            return

        loop = asyncio.get_event_loop()
        task = loop.create_task(self._user_callback(button, *self._user_args, **self._user_kwargs))
        task.add_done_callback(safe_task_callback)

    def connect(self, signal: str, callback: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        if signal != "clicked":
            super().connect(signal)
        else:
            self._user_callback = callback
            self._user_args = args
            self._user_kwargs = kwargs

    def emit(self, signal: str) -> None:
        if signal != "clicked":
            super().emit(signal)
        else:
            self.__callback(self)


class StatusBar(Gtk.Grid):
    def __init__(self) -> None:
        super().__init__()
        _separator = Gtk.Separator()
        self.attach(_separator, 0, 0, 1, 1)

        self._status = Gtk.Label()
        self._status.set_hexpand(True)
        self.attach(self._status, 0, 1, 1, 1)

        self.messages = {}
        for module in config.plugins.keys():
            self.messages[module] = {"warning": "", "critical": ""}

        loop = asyncio.get_event_loop()
        task = loop.create_task(self.__loop_messages())
        task.add_done_callback(safe_task_callback)

    async def __loop_messages(self) -> None:
        while True:
            # when query is empty
            if all(all(value == '' for value in messages.values()) for messages in self.messages.values()):
                self._status.set_css_classes([])
                self._status.set_text('')
                await asyncio.sleep(1)
                continue

            for module_name, module_messages in self.messages.items():
                for level in ["warning", "critical"]:
                    if module_messages[level]:
                        self._status.set_css_classes([level])
                        self._status.set_text(f"{module_name}: {module_messages[level]}")
                        await asyncio.sleep(3)

    def set_warning(self, module: str, message: str) -> None:
        self.messages[module]["warning"] = message

    def set_critical(self, module: str, message: str) -> None:
        self.messages[module]["critical"] = message

    def clear(self, module: str) -> None:
        self.messages[module] = {"warning": "", "critical": ""}


class SimpleTextTree(Gtk.Grid):
    def __init__(
            self,
            *elements: str,
            overlay_scrolling: bool = True,
            resizable: bool = True,
            fixed_width: int = 0,
            model: Callable[..., Union[Gtk.TreeStore, Gtk.ListStore]] = Gtk.TreeStore,
    ) -> None:
        super().__init__()
        self._scrolled_window = Gtk.ScrolledWindow()
        self._scrolled_window.set_overlay_scrolling(overlay_scrolling)
        self.attach(self._scrolled_window, 0, 0, 1, 1)

        # noinspection PyUnusedLocal
        self._store = model(*[str for header in elements])
        self._view = Gtk.TreeView()
        self._view.set_model(self._store)
        self._view.set_hexpand(True)
        self._view.set_vexpand(True)
        self._scrolled_window.set_child(self._view)

        self._lock = False
        self._lock_label = Gtk.Label()
        self._lock_label.hide()
        self._lock_label.set_vexpand(True)
        self._lock_label.set_hexpand(True)

        self._lock_label.set_markup(
            markup(
                _("Waiting another process"),
                background="#0000FF77",
                color="white",
                font_size="xx-large",
            )
        )

        self.attach(self._lock_label, 0, 0, 1, 1)

        self._disabled = False
        self._disabled_label = Gtk.Label()
        self._disabled_label.hide()
        self._disabled_label.set_vexpand(True)
        self._disabled_label.set_hexpand(True)

        self._disabled_label.set_markup(
            markup(
                _("Disabled"),
                background="#FF000077",
                color="white",
                font_size="xx-large",
            )
        )

        self.attach(self._disabled_label, 0, 0, 1, 1)

        renderer = Gtk.CellRendererText()

        for index, header in enumerate(elements):
            column = Gtk.TreeViewColumn(header, renderer, markup=index)
            column.set_resizable(resizable)

            if fixed_width:
                column.set_fixed_width(fixed_width)

            self._view.append_column(column)

    async def wait_available(self) -> None:
        while self.lock:
            await asyncio.sleep(1)

    @property
    def lock(self) -> bool:
        return self._lock

    @lock.setter
    def lock(self, enable_lock: bool) -> None:
        if enable_lock:
            log.debug(_("Waiting another process"))
            self.set_focusable(False)
            self.set_sensitive(False)
            self._lock_label.show()
            self._lock = True
        else:
            self.set_focusable(True)
            self.set_sensitive(True)
            self._lock_label.hide()
            self._lock = False

    @property
    def disabled(self) -> bool:
        return self._disabled

    @disabled.setter
    def disabled(self, disabled: bool) -> None:
        if disabled:
            self.set_focusable(False)
            self.set_sensitive(False)
            self._disabled_label.show()
            self._disabled = True
        else:
            self.set_focusable(True)
            self.set_sensitive(True)
            self._disabled_label.hide()
            self._disabled = False

    @property
    def store(self) -> Union[Gtk.TreeStore, Gtk.ListStore]:
        return self._store

    @property
    def view(self) -> Gtk.TreeView:
        return self._view


class SimpleStatus(Gtk.Frame):
    def __init__(self) -> None:
        super().__init__()
        self._style_context = self.get_style_context()
        self._style_provider = Gtk.CssProvider()
        self._style_provider.load_from_data(b"frame { background-color: black; }")
        self._style_context.add_provider(self._style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self._grid = Gtk.Grid()
        self._grid.set_margin_start(10)
        self._grid.set_margin_end(10)
        self._grid.set_margin_top(10)
        self._grid.set_margin_bottom(10)
        self.set_child(self._grid)

        self._label = Gtk.Label()
        self._label.set_halign(Gtk.Align.START)
        self._grid.attach(self._label, 0, 0, 1, 1)

        self.info(_("Waiting"))
        self.set_size_request(100, 60)

    def error(self, text: str) -> None:
        self._label.set_markup(markup(text, color='hotpink', face='monospace'))

    def info(self, text: str) -> None:
        self._label.set_markup(markup(text, color='cyan', face='monospace'))


def when_running(function: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(function)
    def wrapper(self: 'Status', *args: Any, **kwargs: Any) -> None:
        if self.play_event.is_set():
            function(self, *args, **kwargs)

    return wrapper


class Status(Gtk.Frame):
    def __init__(self, display_size: int) -> None:
        super().__init__()

        # noinspection PyUnusedLocal
        self._default_display_text = ' '.join(['_' for n in range(1, display_size)])
        self._gtk_settings = Gtk.Settings.get_default()
        self.set_margin_top(0)
        self.set_margin_bottom(0)

        self._grid = Gtk.Grid()
        self._grid.set_margin_top(10)
        self._grid.set_margin_bottom(0)
        self.set_child(self._grid)

        self._display = ClickableLabel()
        self._display.set_markup(markup(self._default_display_text, font_size='large', font_weight='bold'))
        self._display.connect("clicked", self.__on_display_event_changed)
        self._display.set_has_tooltip(True)
        self._grid.attach(self._display, 0, 0, 1, 1)

        self._play_pause_button = Gtk.ToggleButton()
        self._play_pause_button.set_margin_end(10)
        self._play_pause_button.set_icon_name("media-playback-start")
        self._play_pause_button.set_can_focus(False)
        self._play_pause_button.connect("toggled", self.__on_play_pause_button_toggled)
        self._grid.attach(self._play_pause_button, 1, 0, 1, 1)

        self._play_event = asyncio.Event()
        self._play_pause_button.emit("clicked")

        self._status = Gtk.Label()
        self._status.set_markup(markup(_("Loading..."), color='green', font_size='small'))
        self._grid.attach(self._status, 0, 1, 1, 1)

        self._info = Gtk.Label()
        self._grid.attach(self._info, 0, 2, 1, 1)

        self._level_bar = Gtk.LevelBar()
        self._grid.attach(self._level_bar, 0, 3, 2, 1)

        self.display = Gdk.Display.get_default()
        self.clipboard = self.display.get_clipboard()

    @property
    def play_event(self) -> asyncio.Event:
        return self._play_event

    @play_event.setter
    def play_event(self, value: bool) -> None:
        if value is True:
            self._play_event.set()
        else:
            self._play_event.clear()

    @staticmethod
    def __sanitize_text(text: str, max_length: int) -> str:
        if len(text) >= max_length:
            return text[:max_length] + '...'

        return text

    def __disable_tooltip(self, event_status: Gtk.Label) -> None:
        self._grid.remove(event_status)
        self._grid.attach(self._status, 0, 1, 1, 1)
        self._display.set_has_tooltip(False)
        self._display.set_sensitive(True)

    def __on_display_event_changed(self, *args: Any, **kwargs: Any) -> None:
        message = _("Text Copied to Clipboard")

        if self._gtk_settings.props.gtk_application_prefer_dark_theme:
            color = 'lightblue'
        else:
            color = 'blue'

        self._display.set_tooltip_text(message)

        event_status = Gtk.Label()
        self._grid.remove(self._status)
        self._grid.attach(event_status, 0, 1, 1, 1)

        event_status.set_markup(markup(message, font_size='small', color=color))
        content = Gdk.ContentProvider.new_for_value(self._display.get_text())
        self.clipboard.set_content(content)
        asyncio.get_event_loop().call_later(5, self.__disable_tooltip, event_status)

        self._display.set_sensitive(False)

    def __on_play_pause_button_toggled(self, button: Gtk.ToggleButton) -> None:
        if button.get_active():
            button.set_icon_name("media-playback-pause")
            self.play_event.set()
        else:
            button.set_icon_name("media-playback-start")
            self.play_event.clear()

    def set_pausable(self, value: bool = True) -> None:
        if value is True:
            self._play_pause_button.show()
        else:
            self._play_pause_button.hide()

    @when_running
    def set_display(self, text: str) -> None:
        text = self.__sanitize_text(text, 25)
        self._display.set_markup(markup(text, font_size='large', font_weight='bold'))

    @when_running
    def set_status(self, text: str) -> None:
        text = self.__sanitize_text(text, 55)
        if self._gtk_settings.props.gtk_application_prefer_dark_theme:
            color = 'lightgreen'
        else:
            color = 'green'

        self._status.set_markup(markup(text, color=color, font_size='small'))

    @when_running
    def set_info(self, text: str) -> None:
        text = self.__sanitize_text(text, 55)
        if self._gtk_settings.props.gtk_application_prefer_dark_theme:
            color = 'lightgreen'
        else:
            color = 'green'

        self._info.set_markup(markup(text, color=color, font_size='small'))

    @when_running
    def set_error(self, text: str) -> None:
        text = self.__sanitize_text(text, 55)
        if self._gtk_settings.props.gtk_application_prefer_dark_theme:
            color = 'hotpink'
        else:
            color = 'red'

        self._status.set_markup(markup(text, color=color, font_size='small'))

    @when_running
    def set_level(self, value: int, max_value: int) -> None:
        self._level_bar.set_value(value)
        self._level_bar.set_max_value(max_value)

    def unset_display(self) -> None:
        self._display.set_markup(markup(self._default_display_text, font_size='large', font_weight='bold'))

    def unset_level(self) -> None:
        self._level_bar.set_value(0)
        self._level_bar.set_max_value(0)


class Section(Gtk.Grid):
    def __init__(self, name: str) -> None:
        super().__init__()
        # for backward compatibility
        self.grid = self

        self.set_name(name)
        self.set_row_spacing(10)
        self.set_column_spacing(10)
        self.set_margin_top(10)
        self.set_margin_bottom(10)

    # noinspection PyProtectedMember
    @staticmethod
    def __get_section_name(item: 'Item') -> str:
        assert isinstance(item._section_name, str)
        return item._section_name

    @staticmethod
    def __show(item: 'Item') -> None:
        item.label.show()
        super(item.__class__, item).show()

    @staticmethod
    def __hide(item: 'Item') -> None:
        item.label.hide()
        super(item.__class__, item).hide()

    def new_item(
            self,
            name: str,
            label: Optional[str],
            widget: Type[Gtk.Widget],
            *grid_position: int,
            items: Optional[OrderedDict[str, str]] = None,
    ) -> Gtk.Widget:
        bases = (widget,)

        body = {
            'label': None,
            '_section_name': None,
            '__init__': lambda item_: super(item_.__class__, item_).__init__(),
            'get_section_name': self.__get_section_name,
            'show': self.__show,
            'hide': self.__hide,
        }

        if label:
            body['label'] = Gtk.Label()

        section = self.get_name()
        option = name
        value: Union[str, bool, int]

        item = type('Item', bases, body)()
        assert isinstance(item, Gtk.Widget)

        item.set_hexpand(True)
        item.set_name(name)
        item._section_name = section

        if item.label:
            item.label.set_name(name)
            item.label.set_text(label)
            item.label.set_halign(Gtk.Align.START)

            self.grid.attach(item.label, *grid_position, 1, 1)
            self.grid.attach_next_to(item, item.label, Gtk.PositionType.RIGHT, 1, 1)
        else:
            self.grid.attach(item, *grid_position, 1, 1)

        if name.startswith('_'):
            return item

        if isinstance(item, Gtk.ComboBoxText):
            assert isinstance(items, OrderedDict), "ComboBox needs items mapping"
            value = config.parser.get(section, option)

            for option_label in items.values():
                item.append_text(_(option_label))

            try:
                current_option = list(items).index(value)
            except ValueError:
                import sys

                error_message = _("Please, fix your config file. Accepted values for {} are:\n{}").format(
                    option,
                    ', '.join(items.keys()),
                )
                log.exception(error_message)
                traceback_info = sys.exc_info()[2]
                fatal_error_dialog(ValueError(error_message), traceback.extract_tb(traceback_info))
                # unset active item
                current_option = -1

            item.set_active(current_option)

        if isinstance(item, Gtk.CheckButton) or isinstance(item, Gtk.Switch):
            value = config.parser.getboolean(section, option)
            item.set_active(value)

        if isinstance(item, Gtk.Entry):
            value = config.parser.get(section, option)

            if value:
                item.set_text(value)

        return item

    def stackup_section(self, text: str, stack: Gtk.Stack, *, scroll: bool = False) -> None:
        name = self.get_name()

        if scroll:
            scroll = Gtk.ScrolledWindow()
            scroll.set_overlay_scrolling(True)
            scroll.set_child(self)
            stack.add_titled(scroll, name, text)
        else:
            stack.add_titled(self, name, text)


def markup(text: str, **kwargs: Any) -> str:
    markup_string = ['<span']

    for key, value in kwargs.items():
        markup_string.append(f'{key}="{value}"')

    markup_string.append(f'>{html.escape(text)}</span>')

    return ' '.join(markup_string)


def unmarkup(text: str) -> str:
    tree = ElementTree.fromstring(text)
    assert isinstance(tree.text, str)
    return tree.text


def copy_childrens(from_model: Gtk.TreeModel, to_model: Gtk.TreeModel, iter_: Gtk.TreeIter, column: int) -> None:
    childrens = from_model.iter_n_children(iter_)

    if childrens:
        for index in range(childrens):
            children_iter = from_model.iter_nth_child(iter_, index)
            value = from_model.get_value(children_iter, column)

            if value:
                to_model.append([value])
            else:
                log.debug(
                    _("Ignoring value from {} on column {} item {} because value is empty").format(
                        children_iter,
                        column,
                        index
                    )
                )
    else:
        value = from_model.get_value(iter_, column)
        to_model.append([value])


def sanitize_confirmation(value: Optional[List[str]]) -> str:
    if not value:
        result = _("Nothing")
    elif len(value) == 1:
        result = value[0]
    else:
        result = _("Various")

    return result


def sanitize_package_details(package_details: List[internals.Package]) -> List[internals.Package]:
    previous: Optional[Tuple[internals.Package, int, int]] = None

    for package in package_details:
        for index, app in enumerate(package.apps):
            if not previous:
                previous = (package, index, app)
                continue

            if previous[2] != app:
                return package_details

    assert isinstance(previous, tuple)
    return [previous[0]]


def remove_letters(text: str) -> str:
    new_text = []

    for char in text:
        if char.isdigit():
            new_text.append(char)

    return ''.join(new_text)


def fatal_error_dialog(
        exception: BaseException,
        stack: Optional[Union[StackSummary, List[FrameType]]] = None,
        transient: Optional[Gtk.Window] = None,
) -> None:
    log.critical("%s: %s", type(exception).__name__, str(exception))

    error_dialog = Gtk.MessageDialog()
    error_dialog.set_transient_for(transient)
    error_dialog.set_title(_("Fatal Error"))
    error_dialog.set_markup(f"{type(exception).__name__}: {str(exception)}")
    error_dialog.set_modal(True)

    message_area = error_dialog.get_message_area()
    secondary_text = Gtk.Label()
    message_area.append(secondary_text)

    if stack:
        log.critical("\n".join([str(frame) for frame in stack]))
        secondary_text.set_text("\n".join([str(frame) for frame in stack]))

    def callback(dialog: Any, _action: Any) -> None:
        loop = asyncio.get_event_loop()
        loop.stop()

        application = Gtk.Application.get_default()

        if application:
            application.quit()

    error_dialog.add_button(_('Ok'), Gtk.ResponseType.OK)
    error_dialog.connect("response", callback)

    error_dialog.show()

    # main application can be not available (like on initialization process)
    if not Gtk.Application.get_default():
        async_gtk.run()


def safe_task_callback(task: asyncio.Task[Any]) -> None:
    if task.cancelled():
        log.debug(_("%s has been stopped due user request"), task.get_coro())
        return

    exception = task.exception()

    if exception and not isinstance(exception, asyncio.CancelledError):
        stack = task.get_stack()
        application = Gtk.Application.get_default()

        fatal_error_dialog(exception, stack, application.get_active_window())


def on_setting_state_set(switch: Gtk.Switch, state: bool) -> None:
    section = switch.get_section_name()
    option = switch.get_name()

    config.new(section, option, state)


def on_setting_changed(entry: Gtk.Entry) -> None:
    current_value = entry.get_text()
    section = entry.get_section_name()
    option = entry.get_name()

    config.new(section, option, current_value)


def on_digit_only_setting_changed(entry: Gtk.Entry) -> None:
    current_value = entry.get_text()
    section = entry.get_section_name()
    option = entry.get_name()

    if current_value.isdigit():
        config.new(section, option, int(current_value))
    else:
        entry.handler_block_by_func(on_digit_only_setting_changed)
        entry.set_text(remove_letters(current_value))
        entry.handler_unblock_by_func(on_digit_only_setting_changed)


def on_combo_setting_changed(combo: Gtk.ComboBoxText, items: OrderedDict[str, str]) -> None:
    current_value = list(items)[combo.get_active()]
    section = combo.get_section_name()
    option = combo.get_name()

    config.new(section, option, current_value)



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#
import aiohttp
import asyncio
import logging
from gi.repository import Gtk, Gdk
from typing import Optional

from stlib import universe, webapi
from . import utils
from .. import i18n, config

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class NewAuthenticatorDialog(Gtk.Dialog):
    def __init__(self, parent_window: Gtk.Window, application: Gtk.Application) -> None:
        super().__init__(use_header_bar=True)
        self.application = application
        self._login_data = None
        self.webapi_session = webapi.SteamWebAPI.get_session(0)

        self.header_bar = self.get_header_bar()

        self.add_authenticator_button = utils.AsyncButton()
        self.add_authenticator_button.set_label(_("Add Authenticator"))
        self.add_authenticator_button.connect("clicked", self.on_add_authenticator_clicked)
        self.header_bar.pack_end(self.add_authenticator_button)

        self.parent_window = parent_window
        self.set_default_size(400, 100)
        self.set_title(_('New Authenticator'))
        self.set_transient_for(parent_window)
        self.set_modal(True)
        self.set_destroy_with_parent(True)
        self.set_resizable(False)

        self.content_area = self.get_content_area()
        self.content_area.set_orientation(Gtk.Orientation.VERTICAL)
        self.content_area.set_spacing(10)
        self.content_area.set_margin_start(10)
        self.content_area.set_margin_end(10)
        self.content_area.set_margin_top(10)
        self.content_area.set_margin_bottom(10)

        self.status = utils.SimpleStatus()
        self.content_area.append(self.status)

        self.user_details_section = utils.Section("Login")
        self.content_area.append(self.user_details_section)

        self.sms_code_item = self.user_details_section.new_item("_sms_code", _("SMS Code:"), Gtk.Entry, 0, 1)

        self.connect('response', lambda dialog, _action: dialog.destroy())

        key_event = Gtk.EventControllerKey()
        key_event.connect('key-released', self.on_key_release_event)
        self.add_controller(key_event)

        self.add_authenticator_button.emit('clicked')

    @property
    def sms_code(self) -> str:
        return self.sms_code_item.get_text()

    @property
    def oauth_token(self) -> str:
        return config.parser.get("login", "oauth_token")

    @property
    def steamid(self) -> Optional[universe.SteamId]:
        steamid = config.parser.getint("login", "steamid")

        if steamid:
            try:
                return universe.generate_steamid(steamid)
            except ValueError:
                log.warning(_("SteamId is invalid"))

        return None

    def on_key_release_event(
            self,
            controller: Gtk.EventControllerKey,
            keyval: int,
            keycode: int,
            state: Gdk.ModifierType,
    ) -> None:
        if keyval == Gdk.KEY_Return:
            self.add_authenticator_button.emit('clicked')

    async def on_add_authenticator_clicked(self, button: Gtk.Button) -> None:
        self.status.info(_("Retrieving user data"))
        button.set_sensitive(False)
        self.user_details_section.hide()
        self.set_size_request(0, 0)

        if not self.oauth_token or not self.steamid:
            self.status.error(_(
                "Some login data is missing. If the problem persists, go to:\n"
                "Settings -> Login -> Advanced -> and click on RESET Everything."
            ))

            return

        if not self._login_data or not self.sms_code:
            try:
                self._login_data = await self.webapi_session.new_authenticator(self.steamid, self.oauth_token)
            except aiohttp.ClientError:
                self.status.error(_("Check your connection. (server down?)"))
            except webapi.AuthenticatorExists:
                self.status.error(_(
                    "There's already an authenticator active for that account.\n"
                    "Remove your current steam authenticator and try again."
                ))
            except webapi.PhoneNotRegistered:
                self.status.error(_(
                    "You must have a phone registered on your steam account to proceed.\n"
                    "Go to your Steam Account Settings, add a Phone Number, and try again."
                ))
            except NotImplementedError as exception:
                import sys
                import traceback
                traceback_info = sys.exc_info()[2]
                utils.fatal_error_dialog(exception, traceback.extract_tb(traceback_info), self.parent_window)
                self.application.on_exit_activate()
            else:
                self.status.info(_("Enter bellow the code received by SMS\nand click on 'Add Authenticator' button"))
                self.user_details_section.show()
                self.sms_code_item.set_text('')
                self.sms_code_item.grab_focus()
            finally:
                button.set_sensitive(True)

            return

        self.status.info(_("Adding authenticator"))

        try:
            await self.webapi_session.add_authenticator(
                self.steamid,
                self.oauth_token,
                self._login_data.auth['shared_secret'],
                self.sms_code,
            )
        except webapi.SMSCodeError:
            self.status.info(_("Invalid SMS Code. Please,\ncheck the code and try again."))
            self.user_details_section.show()
            self.sms_code_item.set_text('')
            self.sms_code_item.grab_focus()
        except aiohttp.ClientError:
            self.status.error(_("Check your connection. (server down?)"))
            self.user_details_section.show()
            self.sms_code_item.set_text('')
            self.sms_code_item.grab_focus()
        except Exception as exception:
            import sys
            import traceback
            traceback_info = sys.exc_info()[2]
            utils.fatal_error_dialog(exception, traceback.extract_tb(traceback_info), self.parent_window)
            self.application.on_exit_activate()
        else:
            self.status.info(_("Saving new secrets"))
            config.new("login", "shared_secret", self._login_data.auth['shared_secret'])
            config.new("login", "identity_secret", self._login_data.auth['identity_secret'])
            config.new("steamguard", "enable", True)
            config.new("confirmations", "enable", True)

            self.status.info(_(
                "RECOVERY CODE\n\n"
                "You will need this code to recovery your Steam Account\n"
                "if you lose access to STNG Authenticator. So, write"
                "down this recovery code.\n\n"
                "YOU WILL NOT ABLE TO VIEW IT AGAIN!\n"
            ))

            revocation_code = self._login_data.auth['revocation_code']

            self.add_authenticator_button.hide()

            revocation_status = utils.Status(6)
            revocation_status.set_pausable(False)
            revocation_status.set_display(revocation_code)
            revocation_status.set_status('')
            self.content_area.append(revocation_status)

            revocation_status.show()

            self.set_deletable(False)

            max_value = 30 * 3
            for offset in range(max_value):
                revocation_status.set_level(offset, max_value)
                await asyncio.sleep(0.3)

            self.set_deletable(True)
        finally:
            button.set_sensitive(True)



#!/usr/bin/env python
#
# Lara Maia <dev@lara.monster> 2015 ~ 2023
#
# The Steam Tools NG is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The Steam Tools NG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#

import logging
from gi.repository import Gtk, Pango
from subprocess import call
from typing import Type

from . import utils
from .. import config, i18n

log = logging.getLogger(__name__)
_ = i18n.get_translation


# noinspection PyUnusedLocal
class SettingsDialog(Gtk.Dialog):
    def __init__(
            self,
            parent_window: Gtk.Window,
            application: Gtk.Application,
    ) -> None:
        super().__init__(use_header_bar=True)
        self.parent_window = parent_window
        self.application = application

        self.set_default_size(300, 150)
        self.set_title(_('Settings'))
        self.set_transient_for(parent_window)
        self.set_modal(True)
        self.set_destroy_with_parent(True)
        self.set_resizable(False)

        self.gtk_settings_class = Gtk.Settings.get_default()

        content_area = self.get_content_area()
        content_grid = Gtk.Grid()
        content_grid.set_row_spacing(10)
        content_grid.set_column_spacing(10)
        content_area.append(content_grid)

        stack = Gtk.Stack()
        stack.set_margin_end(10)
        content_grid.attach(stack, 1, 0, 1, 1)

        sidebar = Gtk.StackSidebar()
        sidebar.set_stack(stack)
        content_grid.attach(sidebar, 0, 0, 1, 1)

        general_section = utils.Section("general")
        general_section.stackup_section(_("General"), stack)

        config_button = Gtk.Button()
        config_button.set_label(_("Config File Directory"))
        config_button.set_name("config_button")
        config_button.set_hexpand(True)
        config_button.connect("clicked", self.on_config_button_clicked)

        if config.parser.get("logger", "log_directory") == str(config.config_file_directory):
            config_button.set_label(_("Config / Log file Directory"))
            general_section.grid.attach(config_button, 0, 1, 2, 1)
        else:
            log_button = Gtk.Button()
            log_button.set_label(_("Log File Directory"))
            log_button.set_name("log_button")
            log_button.connect("clicked", self.on_log_button_clicked)
            general_section.grid.attach(config_button, 0, 1, 1, 1)
            general_section.grid.attach(log_button, 1, 1, 1, 1)

        theme = general_section.new_item(
            "theme",
            _("Theme:"),
            Gtk.ComboBoxText,
            0, 3,
            items=config.gtk_themes,
        )

        theme.connect('changed', self.on_theme_changed)

        language_item = general_section.new_item(
            "language", _("Language"), Gtk.ComboBoxText, 0, 4, items=config.translations
        )
        language_item.connect("changed", self.update_language)

        show_close_button = general_section.new_item(
            "show_close_button",
            _("Show close button:"),
            Gtk.Switch,
            0, 5,
        )
        show_close_button.set_halign(Gtk.Align.END)

        show_close_button.connect('state-set', self.on_show_close_button_state_set)

        logger_section = utils.Section("logger")
        logger_section.stackup_section(_("Logger"), stack)

        log_level_item = logger_section.new_item(
            "log_level",
            _("Level:"),
            Gtk.ComboBoxText,
            0, 0,
            items=config.log_levels,
        )

        log_console_level_item = logger_section.new_item(
            "log_console_level",
            _("Console level:"),
            Gtk.ComboBoxText,
            0, 1,
            items=config.log_levels,
        )

        log_level_item.connect("changed", utils.on_combo_setting_changed, config.log_levels)
        log_console_level_item.connect("changed", utils.on_combo_setting_changed, config.log_levels)

        log_color = logger_section.new_item("log_color", _("Log color:"), Gtk.Switch, 0, 2)
        log_color.set_halign(Gtk.Align.END)
        log_color.connect('state-set', utils.on_setting_state_set)

        self.connect('response', lambda dialog, response_id: self.destroy())

    @staticmethod
    def on_log_button_clicked(button: Gtk.Button) -> None:
        call(f'{config.file_manager} {config.parser.get("logger", "log_directory")}')

    @staticmethod
    def on_config_button_clicked(button: Gtk.Button) -> None:
        call(f'{config.file_manager} {str(config.config_file_directory)}')

    def on_show_close_button_state_set(self, switch: Gtk.Switch, state: bool) -> None:
        if state:
            self.parent_window.set_deletable(True)
        else:
            self.parent_window.set_deletable(False)

        config.new('general', 'show_close_button', state)

    def on_theme_changed(self, combo: Gtk.ComboBoxText) -> None:
        theme = list(config.gtk_themes)[combo.get_active()]

        if theme == 'dark':
            self.gtk_settings_class.props.gtk_application_prefer_dark_theme = True
        else:
            self.gtk_settings_class.props.gtk_application_prefer_dark_theme = False

        config.new('general', 'theme', theme)

    def update_language(self, combo: Gtk.ComboBoxText) -> None:
        language = list(config.translations)[combo.get_active()]
        config.new('general', 'language', language)
        refresh_widget_childrens(self)
        refresh_widget_childrens(self.parent_window)


def refresh_widget_childrens(widget: Type[Gtk.Widget]) -> None:
    next_child = widget.get_first_child()

    while next_child:
        refresh_widget(next_child)
        next_child = next_child.get_next_sibling()


def refresh_widget(widget: Type[Gtk.Widget]) -> None:
    if isinstance(widget, Gtk.MenuButton):
        refresh_widget(widget.get_popover())
        return

    if isinstance(widget, Gtk.ComboBoxText):
        model = widget.get_model()

        for index, row in enumerate(model):
            combo_item_label = model.get_value(row.iter, 0)

            try:
                cached_text = i18n.cache[i18n.new_hash(combo_item_label)]
            except KeyError:
                log.debug("it's not an i18n string: %s", combo_item_label)
                return

            model.set_value(row.iter, 0, _(cached_text))

        log.debug('ComboBox refreshed: %s', widget)

    if isinstance(widget, Gtk.Label):
        try:
            cached_text = i18n.cache[i18n.new_hash(widget.get_text())]
        except KeyError:
            log.debug("it's not an i18n string: %s", widget.get_text())
            return

        if widget.get_use_markup():
            old_attributes = Pango.Layout.get_attributes(widget.get_layout())
            widget.set_text(_(cached_text))
            widget.set_attributes(old_attributes)
        else:
            widget.set_text(_(cached_text))

        log.debug('widget refreshed: %s', str(widget))
        return

    log.debug('widget not refresh: %s', str(widget))
    refresh_widget_childrens(widget)
